from abc import ABC, abstractmethod
from typing import Any, BinaryIO, Optional

from .._pointappender import IPointAppender
from .._pointreader import IPointReader
from .._pointwriter import IPointWriter
from ..header import LasHeader
from .selection import DecompressionSelection


class ILazBackend(ABC):
    @abstractmethod
    def is_available(self) -> bool: ...

    @property
    @abstractmethod
    def supports_append(self) -> bool: ...

    @abstractmethod
    def create_appender(self, dest: BinaryIO, header: LasHeader) -> IPointAppender: ...

    @abstractmethod
    def create_reader(
        self,
        source: Any,
        header: LasHeader,
        decompression_selection: Optional[DecompressionSelection] = None,
    ) -> IPointReader: ...

    @abstractmethod
    def create_writer(
        self,
        dest: Any,
        header: LasHeader,
    ) -> IPointWriter: ...



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

import numpy as np
import typer

import laspy

from ..copc import Bounds, CopcReader

app = typer.Typer(help="COPC related commands")


@dataclass
class Level:
    inner: Union[int, range]

    @classmethod
    def parse(cls, value_string: str) -> "Level":
        if ".." in value_string:
            # range rust-style
            start_str, end_str = value_string.split("..")
            start = int(start_str)
            if end_str.startswith("="):
                end = int(end_str[1:]) + 1
            else:
                end = int(end_str)

            inner = range(start, end)
        elif ":" in value_string:
            # range python style
            start, end = map(int, value_string.split(":"))
            inner = range(start, end)
        else:
            inner = int(value_string)

        return cls(inner)


def parse_bounds(value_string: str) -> Bounds:
    first_closing_bracket = value_string.find("]")
    second_opening_bracket = value_string[first_closing_bracket:].find("[")

    mins_str = value_string[1:first_closing_bracket]
    maxes_str = value_string[first_closing_bracket + second_opening_bracket + 1 : -1]

    mins = np.array(mins_str.split(","), dtype=np.float64)
    maxes = np.array(maxes_str.split(","), dtype=np.float64)

    return Bounds(mins, maxes)


@app.command()
def query(
    source: str = typer.Argument(
        ..., help="The COPC ressource, can be a path or a url"
    ),
    dest: Path = typer.Option(
        ...,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    bounds: Optional[Bounds] = typer.Option(
        None,
        parser=parse_bounds,
        help="""
                The bounds for which you wish to aquire points.
                If None, the whole file's bounds will be considered
                2D bounds are suported, (No point will be filtered on its Z coordinate)
""",
    ),
    resolution: Optional[float] = typer.Option(
        None,
        help=""" Limits the octree levels to be queried in order to have
                a point cloud with the requested resolution.

                - The unit is the one of the data.

                - If None, the resulting cloud will be at the
                  full resolution offered by the COPC source

                - Mutually exclusive with level parameter
""",
    ),
    level: Optional[Level] = typer.Option(
        None,
        parser=Level.parse,
        help=""" Level of detail (LOD)
                By default all levels are considered.

               - If it is an int, only points that are of the requested LOD
                 will be returned.
               - If it is a range, points for which the LOD is within the range
                 will be returned
""",
    ),
    safe_guard: bool = typer.Option(
        True,
        help="""
            Enable / Disable the safeguard that prevents downloading
            the whole COPC file when the option was provided
""",
    ),
):
    """
    Query the COPC file to retrieve the points matching the
    requested bounds and level/resolution


    Examples:



    laspy COPC "https://s3.amazonaws.com/hobu-lidar/melbourne-2018.copc.laz"
        output.laz
        --bounds "[3.1e+05, 5.8+06, -2.0e+01], [3.17e+05, 5.81e+06, 3.0e+02]"

    ---


    laspy COPC "https://s3.amazonaws.com/hobu-lidar/melbourne-2018.copc.laz"
        output.laz
        --bounds "[3.1e+05, 5.8+06], [3.17e+05, 5.81e+06]"

    ---

    laspy COPC "https://s3.amazonaws.com/hobu-lidar/melbourne-2018.copc.laz"
        output.laz
        --bounds "[3.1e+05, 5.8+06], [3.17e+05, 5.81e+06]"
        --resolution 2

    ---

    laspy COPC "https://s3.amazonaws.com/hobu-lidar/melbourne-2018.copc.laz"
        output.laz
        --bounds "[3.1e+05, 5.8+06], [3.17e+05, 5.81e+06]"
        --level 1

    ---

    # Level can be a range (here [0, 2[)

    laspy COPC "https://s3.amazonaws.com/hobu-lidar/melbourne-2018.copc.laz"
        output.laz
        --bounds "[3.1e+05, 5.8+06], [3.17e+05, 5.81e+06]"
        --level 0:2

    ---

    # Level can be a range (here [0, 2[)

    laspy COPC "https://s3.amazonaws.com/hobu-lidar/melbourne-2018.copc.laz"
        output.laz
        --bounds "[3.1e+05, 5.8+06], [3.17e+05, 5.81e+06]"
        --level 0..2

    laspy COPC "https://s3.amazonaws.com/hobu-lidar/melbourne-2018.copc.laz"
        output.laz
        --bounds "[3.1e+05, 5.8+06], [3.17e+05, 5.81e+06]"
        --level 0..=2
    """

    level = None if level is None else level.inner

    with CopcReader.open(source) as crdr:
        if safe_guard is True and all(
            map(lambda x: x is None, (level, bounds, resolution))
        ):
            print(
                "You did not specify any of --bounds, --level or --resolution options"
            )
            print(
                f"The command is going to download the whole COPC resource ({crdr.header.point_count}) points"
            )
            anwser = input("Do you really want to continue ? [y/N]")
            if anwser not in ("y", "Y", "yes"):
                return

        points = crdr.query(bounds=bounds, resolution=resolution, level=level)
        pct = len(points) / crdr.header.point_count * 100
        print(
            f"{len(points)} points downloaded\n{pct}% of the total number of points in COPC resource"
        )

        new_header = laspy.LasHeader(
            version=crdr.header.version, point_format=crdr.header.point_format
        )
        new_header.offsets = crdr.header.offsets
        new_header.scales = crdr.header.scales

        with laspy.open(dest, mode="w", header=new_header) as f:
            f.write_points(points)



from typing import BinaryIO, Iterable, Literal, Optional, Union, overload

from . import LasWriter, PointFormat
from .compression import DecompressionSelection, LazBackend
from .header import LasHeader
from .lasappender import LasAppender
from .lasdata import LasData
from .lasmmap import LasMMAP
from .lasreader import LasReader
from .typehints import PathLike

LazBackend = LazBackend
DecompressionSelection = DecompressionSelection

@overload
def open_las(
    source: PathLike,
    mode: Literal["r"] = ...,
    laz_backend: Union[LazBackend, Iterable[LazBackend]] = ...,
) -> LasReader: ...
@overload
def open_las(
    source: BinaryIO,
    mode: Literal["r"] = ...,
    closefd: bool = ...,
    laz_backend: Union[LazBackend, Iterable[LazBackend]] = ...,
) -> LasReader: ...
@overload
def open_las(
    source: PathLike,
    mode: Literal["w"],
    header: LasHeader,
    do_compress: Optional[bool] = ...,
    laz_backend: Union[LazBackend, Iterable[LazBackend]] = ...,
) -> LasWriter: ...
@overload
def open_las(
    source: BinaryIO,
    mode: Literal["w"],
    header: LasHeader,
    do_compress: Optional[bool] = ...,
    closefd: bool = ...,
    laz_backend: Union[LazBackend, Iterable[LazBackend]] = ...,
) -> LasWriter: ...
@overload
def open_las(
    source: PathLike,
    mode: Literal["a"],
    laz_backend: Union[LazBackend, Iterable[LazBackend]] = ...,
) -> LasAppender: ...
@overload
def open_las(
    source: BinaryIO,
    mode: Literal["a"],
    closefd: bool = ...,
    laz_backend: Union[LazBackend, Iterable[LazBackend]] = ...,
) -> LasAppender: ...
def read_las(
    source: Union[BinaryIO, PathLike],
    closefd: bool = True,
    laz_backend: Union[
        LazBackend, Iterable[LazBackend]
    ] = LazBackend.detect_available(),
    decompression_selection: DecompressionSelection = DecompressionSelection.all(),
) -> LasData: ...
def mmap_las(filename: PathLike) -> LasMMAP: ...
def merge_las(las_files: Union[Iterable[LasData], LasData]) -> LasData: ...
def create_las(
    *, point_format: Union[int, PointFormat] = 0, file_version: Optional[str] = 0
) -> LasData: ...
def convert(
    source_las: LasData,
    *,
    point_format_id: Optional[int] = ...,
    file_version: Optional[str] = ...,
) -> LasData: ...
def create_from_header(header: LasHeader) -> LasData: ...
def write_then_read_again(
    las: LasData, do_compress: bool = ..., laz_backend: LazBackend = ...
) -> LasData: ...



import enum
import io
import logging
import struct
import typing
from copy import deepcopy
from datetime import date, timedelta
from typing import BinaryIO, Iterable, List, NamedTuple, Optional, Union
from uuid import UUID

import numpy as np

from . import __version__, extradims
from ._compression.format import (
    compressed_id_to_uncompressed,
    is_point_format_compressed,
    uncompressed_id_to_compressed,
)
from .errors import LaspyException
from .point import dims
from .point.format import ExtraBytesParams, PointFormat
from .point.record import PackedPointRecord
from .utils import read_string, write_string
from .vlrs import VLR
from .vlrs.geotiff import create_geotiff_projection_vlrs
from .vlrs.known import (
    ExtraBytesStruct,
    ExtraBytesVlr,
    GeoAsciiParamsVlr,
    GeoKeyDirectoryVlr,
    WktCoordinateSystemVlr,
)
from .vlrs.vlrlist import VLRList

logger = logging.getLogger(__name__)

GENERATING_SOFTWARE_LEN = 32
SYSTEM_IDENTIFIER_LEN = 32

LAS_FILE_SIGNATURE = b"LASF"
DEFAULT_GENERATING_SOFTWARE = f"laspy {__version__}"
assert len(DEFAULT_GENERATING_SOFTWARE) < GENERATING_SOFTWARE_LEN


class Version(NamedTuple):
    major: int
    minor: int

    @classmethod
    def from_str(cls, string: str) -> "Version":
        major, minor = tuple(map(int, string.split(".")))
        return cls(major, minor)

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        else:
            return other.major == self.major and other.minor == self.minor

    def __str__(self):
        return f"{self.major}.{self.minor}"


class GpsTimeType(enum.IntEnum):
    WEEK_TIME = 0
    STANDARD = 1


class GlobalEncoding:
    GPS_TIME_TYPE_MASK = 0b0000_0000_0000_0001
    WAVEFORM_INTERNAL_MASK = 0b0000_0000_0000_0010  # 1.3
    WAVEFORM_EXTERNAL_MASK = 0b0000_0000_0000_0100  # 1.3
    SYNTHETIC_RETURN_NUMBERS_MASK = 0b0000_0000_0000_1000  # 1.3
    WKT_MASK = 0b0000_0000_0001_0000  # 1.4
    RESERVED = 0b0000_0000_0010_0000
    GPS_TIME_OFFSET_MASK = 0b0000_0000_0100_0000

    def __init__(self, value=0):
        self.value = value

    def _set_bit(self, mask):
        self.value |= mask

    def _unset_bit(self, mask):
        self.value &= ~mask

    def _set_if_true(self, mask, value):
        if bool(value) is True:
            self._set_bit(mask)
        else:
            self._unset_bit(mask)

    @property
    def gps_time_type(self) -> GpsTimeType:
        return GpsTimeType(self.value & self.GPS_TIME_TYPE_MASK)

    @gps_time_type.setter
    def gps_time_type(self, value: GpsTimeType):
        self.value &= ~self.GPS_TIME_TYPE_MASK
        self.value |= int(value) & self.GPS_TIME_TYPE_MASK

    @property
    def waveform_data_packets_internal(self) -> bool:
        return bool(self.value & self.WAVEFORM_INTERNAL_MASK)

    @waveform_data_packets_internal.setter
    def waveform_data_packets_internal(self, value):
        self._set_if_true(self.WAVEFORM_INTERNAL_MASK, value)

    @property
    def waveform_data_packets_external(self) -> bool:
        return bool(self.value & self.WAVEFORM_EXTERNAL_MASK)

    @waveform_data_packets_external.setter
    def waveform_data_packets_external(self, value):
        self._set_if_true(self.WAVEFORM_EXTERNAL_MASK, value)

    @property
    def synthetic_return_numbers(self) -> bool:
        return bool(self.value & self.SYNTHETIC_RETURN_NUMBERS_MASK)

    @synthetic_return_numbers.setter
    def synthetic_return_numbers(self, value):
        self._set_if_true(self.SYNTHETIC_RETURN_NUMBERS_MASK, value)

    @property
    def wkt(self) -> bool:
        return bool(self.value & self.WKT_MASK)

    @wkt.setter
    def wkt(self, value):
        self._set_if_true(self.WKT_MASK, value)

    @property
    def gps_time_offset(self) -> bool:
        return bool(self.value & self.GPS_TIME_OFFSET_MASK)

    @gps_time_offset.setter
    def gps_time_offset(self, value):
        self._set_if_true(self.GPS_TIME_OFFSET_MASK, value)
        if value:
            # Standard says that of the offset mask is true,
            # then the gps time type must also be
            self._set_if_true(self.GPS_TIME_TYPE_MASK, value)

    @classmethod
    def read_from(cls, stream: BinaryIO) -> "GlobalEncoding":
        return cls(int.from_bytes(stream.read(2), byteorder="little", signed=False))

    def write_to(self, stream: BinaryIO) -> None:
        stream.write(self.value.to_bytes(2, byteorder="little", signed=False))


class LasHeader:
    """Contains the information from the header of as LAS file
    with 'implementation' field left out and 'users' field
    stored in more ergonomic classes.

    This header also contains the VLRs

    Examples
    --------

    Creating a default header:

    >>> header = LasHeader()
    >>> header
    <LasHeader(1.2, <PointFormat(3, 0 bytes of extra dims)>)>

    Creating a header with the wanted version and point format:

    >>> header = LasHeader(version=Version(1, 4), point_format=PointFormat(6))
    >>> header
    <LasHeader(1.4, <PointFormat(6, 0 bytes of extra dims)>)>

    >>> header = LasHeader(version="1.4", point_format=6)
    >>> header
    <LasHeader(1.4, <PointFormat(6, 0 bytes of extra dims)>)>
    """

    #: The default version used when None is given to init
    DEFAULT_VERSION = Version(1, 2)
    #: The default point format Used when None is given to init
    DEFAULT_POINT_FORMAT = PointFormat(3)

    _OLD_LASPY_NAMES = {
        "max": "maxs",
        "min": "mins",
        "scale": "scales",
        "offset": "offsets",
        "filesource_id": "file_source_id",
        "system_id": "system_identifier",
        "date": "creation_date",
        "point_return_count": "number_of_points_by_return",
        "software_id": "generating_software",
        "point_records_count": "point_count",
    }

    def __init__(
        self,
        *,
        version: Optional[Union[Version, str]] = None,
        point_format: Optional[Union[PointFormat, int]] = None,
    ) -> None:
        if isinstance(point_format, int):
            point_format = PointFormat(point_format)
        if isinstance(version, str):
            version = Version.from_str(version)

        if version is None and point_format is None:
            version = LasHeader.DEFAULT_VERSION
            point_format = deepcopy(LasHeader.DEFAULT_POINT_FORMAT)
        elif version is not None and point_format is None:
            point_format = PointFormat(dims.min_point_format_for_version(str(version)))
        elif version is None and point_format is not None:
            version = Version.from_str(
                dims.preferred_file_version_for_point_format(point_format.id)
            )
        dims.raise_if_version_not_compatible_with_fmt(point_format.id, str(version))

        #: File source id
        self.file_source_id: int = 0
        self.global_encoding: GlobalEncoding = GlobalEncoding()
        #: Project ID
        #: Initialized to null UUID
        self.uuid: UUID = UUID(bytes_le=b"\0" * 16)
        self._version: Version = version
        #: System identifier
        #: Initialized to 'OTHER'
        self.system_identifier: Union[str, bytes] = "OTHER"
        #: The software which generated the file
        #: Initialized to 'laspy'
        self.generating_software: Union[str, bytes] = DEFAULT_GENERATING_SOFTWARE
        self._point_format: PointFormat = point_format
        #: Day the file was created, initialized to date.today
        self.creation_date: Optional[date] = date.today()
        #: The number of points in the file
        self.point_count: int = 0
        #: The numbers used to scale the x,y,z coordinates
        self.scales: np.ndarray = np.array([0.01, 0.01, 0.01], dtype=np.float64)
        #: The numbers used to offset the x,y,z coordinates
        self.offsets: np.ndarray = np.zeros(3, dtype=np.float64)
        # The max values for x,y,z
        self.maxs: np.ndarray = np.zeros(3, dtype=np.float64)
        # The min values for x,y,z
        self.mins: np.ndarray = np.zeros(3, dtype=np.float64)

        #: Number of points by return
        #: for las <= 1.2 only the first 5 elements matters
        self.number_of_points_by_return: np.ndarray = np.zeros(15, dtype=np.uint64)

        #: The VLRS
        self._vlrs: VLRList = VLRList()

        #: Extra bytes between end of header and first vlrs
        self.extra_header_bytes: bytes = b""
        #: Extra bytes between end of vlr end first point
        self.extra_vlr_bytes: bytes = b""

        #: Las >= 1.3
        self.start_of_waveform_data_packet_record: int = 0

        #: Las >= 1.4
        #: Offset to the first evlr in the file
        self.start_of_first_evlr: int = 0
        #: The number of evlrs in the file
        self.number_of_evlrs: int = 0

        #: Las >= 1.5
        #: Max value of GPS Time
        self.max_gps_time: float = 0.0
        #: Min value of GPS Time
        self.min_gps_time: float = 0.0
        #: Used to optimize GPS Time precision
        self.gps_time_offset: int = 0

        #: EVLRs, even though they are not stored in the 'header'
        #: part of the file we keep them in this class
        #: as they contain same information as vlr.
        #: None when the file does not support EVLR
        self.evlrs: Optional[VLRList] = None

        # Info we keep because it's useful for us but not the user
        self.offset_to_point_data: int = 0
        self.are_points_compressed: bool = False

        self._sync_extra_bytes_vlr()

    @property
    def point_format(self) -> PointFormat:
        """The point format"""
        return self._point_format

    @point_format.setter
    def point_format(self, new_point_format: PointFormat) -> None:
        dims.raise_if_version_not_compatible_with_fmt(
            new_point_format.id, str(self.version)
        )
        self._point_format = new_point_format
        self._sync_extra_bytes_vlr()

    @property
    def version(self) -> Version:
        """The version"""
        return self._version

    @version.setter
    def version(self, version: Version) -> None:
        dims.raise_if_version_not_compatible_with_fmt(
            self.point_format.id, str(version)
        )
        self._version = version

    # scale properties
    @property
    def x_scale(self) -> float:
        return self.scales[0]

    @x_scale.setter
    def x_scale(self, value: float) -> None:
        self.scales[0] = value

    @property
    def y_scale(self) -> float:
        return self.scales[1]

    @y_scale.setter
    def y_scale(self, value: float) -> None:
        self.scales[1] = value

    @property
    def z_scale(self) -> float:
        return self.scales[2]

    @z_scale.setter
    def z_scale(self, value: float) -> None:
        self.scales[2] = value

    # offset properties
    @property
    def x_offset(self) -> float:
        return self.offsets[0]

    @x_offset.setter
    def x_offset(self, value: float) -> None:
        self.offsets[0] = value

    @property
    def y_offset(self) -> float:
        return self.offsets[1]

    @y_offset.setter
    def y_offset(self, value: float) -> None:
        self.offsets[1] = value

    @property
    def z_offset(self) -> float:
        return self.offsets[2]

    @z_offset.setter
    def z_offset(self, value: float) -> None:
        self.offsets[2] = value

    # max properties
    @property
    def x_max(self) -> float:
        return self.maxs[0]

    @x_max.setter
    def x_max(self, value: float) -> None:
        self.maxs[0] = value

    @property
    def y_max(self) -> float:
        return self.maxs[1]

    @y_max.setter
    def y_max(self, value: float) -> None:
        self.maxs[1] = value

    @property
    def z_max(self) -> float:
        return self.maxs[2]

    @z_max.setter
    def z_max(self, value: float) -> None:
        self.maxs[2] = value

    # min properties
    @property
    def x_min(self) -> float:
        return self.mins[0]

    @x_min.setter
    def x_min(self, value: float) -> None:
        self.mins[0] = value

    @property
    def y_min(self) -> float:
        return self.mins[1]

    @y_min.setter
    def y_min(self, value: float) -> None:
        self.mins[1] = value

    @property
    def z_min(self) -> float:
        return self.mins[2]

    @z_min.setter
    def z_min(self, value: float) -> None:
        self.mins[2] = value

    @property
    def vlrs(self) -> VLRList:
        return self._vlrs

    @vlrs.setter
    def vlrs(self, vlrs: typing.Iterable[VLR]) -> None:
        self._vlrs = VLRList(vlrs)

        try:
            self.vlrs.extract("LaszipVlr")
        except ValueError:
            pass

        self._sync_extra_bytes_vlr()

    def add_extra_dims(self, params: List[ExtraBytesParams]) -> None:
        for param in params:
            self.point_format.add_extra_dimension(param)
        self._sync_extra_bytes_vlr()

    def add_extra_dim(self, params: ExtraBytesParams):
        self.add_extra_dims([params])

    def add_crs(self, crs: "pyproj.CRS", keep_compatibility: bool = True) -> None:
        """Add a Coordinate Reference System VLR from a pyproj CRS object.

        The type of VLR created depends on the las version and point format
        version. Las version >= 1.4 use WKT string, las version < 1.4 and point
        format < 6 use GeoTiff tags.

        .. warning::
            This requires `pyproj`.

        .. warning::
            Not all CRS are supported when adding GeoTiff tags.
            For example, custom CRS.

            Typically, if the CRS has an EPSG code it will be supported.
        """
        import pyproj

        # check and remove any existing crs vlrs
        for crs_vlr_name in (
            "WktCoordinateSystemVlr",
            "GeoKeyDirectoryVlr",
            "GeoAsciiParamsVlr",
            "GeoDoubleParamsVlr",
        ):
            try:
                self._vlrs.extract(crs_vlr_name)
            except IndexError:
                pass

        new_ver = self._version >= Version(1, 4)
        new_pt = self.point_format.id >= 6

        if new_pt or (new_ver and not keep_compatibility):
            self._vlrs.append(WktCoordinateSystemVlr(crs.to_wkt()))
            self.global_encoding.wkt = True
        else:
            self._vlrs.extend(create_geotiff_projection_vlrs(crs))

    def remove_extra_dim(self, name: str) -> None:
        self.remove_extra_dims([name])

    def remove_extra_dims(self, names: Iterable[str]) -> None:
        for name in names:
            self.point_format.remove_extra_dimension(name)
        self._sync_extra_bytes_vlr()

    def set_version_and_point_format(
        self, version: Version, point_format: PointFormat
    ) -> None:
        dims.raise_if_version_not_compatible_with_fmt(point_format.id, str(version))
        self._version = version
        self.point_format = point_format

    def partial_reset(self) -> None:
        f64info = np.finfo(np.float64)
        self.maxs = np.ones(3, dtype=np.float64) * f64info.min
        self.mins = np.ones(3, dtype=np.float64) * f64info.max

        self.max_gps_time = float(f64info.min)
        self.min_gps_time = float(f64info.max)

        self.start_of_first_evlr = 0
        self.number_of_evlrs = 0
        self.point_count = 0
        self.number_of_points_by_return = np.zeros(15, dtype=np.uint64)

        try:
            eb_vlr = self.vlrs.get("ExtraBytesVlr")[0]
            eb_vlr.partial_reset()
        except IndexError:
            pass

    def update(self, points: PackedPointRecord) -> None:
        self.partial_reset()
        if not points:
            self.maxs = [0.0, 0.0, 0.0]
            self.mins = [0.0, 0.0, 0.0]
        else:
            self.grow(points)

    def grow(self, points: PackedPointRecord) -> None:
        self.x_max = max(
            self.x_max,
            (points["X"].max() * self.x_scale) + self.x_offset,
        )
        self.y_max = max(
            self.y_max,
            (points["Y"].max() * self.y_scale) + self.y_offset,
        )
        self.z_max = max(
            self.z_max,
            (points["Z"].max() * self.z_scale) + self.z_offset,
        )
        self.x_min = min(
            self.x_min,
            (points["X"].min() * self.x_scale) + self.x_offset,
        )
        self.y_min = min(
            self.y_min,
            (points["Y"].min() * self.y_scale) + self.y_offset,
        )
        self.z_min = min(
            self.z_min,
            (points["Z"].min() * self.z_scale) + self.z_offset,
        )

        try:
            gps_time = points["gps_time"]
            self.max_gps_time = max(self.max_gps_time, gps_time.max())
            self.min_gps_time = min(self.min_gps_time, gps_time.min())
        except ValueError:
            pass

        for return_number, count in zip(
            *np.unique(points.return_number, return_counts=True)
        ):
            if return_number == 0:
                continue
            if return_number > len(self.number_of_points_by_return):
                break  # np.unique sorts unique values
            self.number_of_points_by_return[return_number - 1] += count
        self.point_count += len(points)

        # grow extra bytes
        vlrs = self.vlrs.get("ExtraBytesVlr")
        if vlrs:
            vlrs[0].grow(points)

    def set_compressed(self, state: bool) -> None:
        self.are_points_compressed = state

    def max_point_count(self) -> int:
        if self.version <= Version(1, 3):
            return np.iinfo(np.uint32).max
        else:
            return np.iinfo(np.uint64).max

    def copy(self):
        return deepcopy(self)

    @classmethod
    def read_from(
        cls, original_stream: BinaryIO, read_evlrs: bool = False
    ) -> "LasHeader":
        """
        Reads the header from the stream

        read_evlrs: If true, evlrs will be read

        Leaves the stream pos right before the point starts
        (regardless of is read_evlrs was true)

        """
        little_endian = "little"
        header = cls()

        stream = io.BytesIO(cls._prefetch_header_data(original_stream))

        file_sig = stream.read(4)
        # This should not be possible as _prefetch already checks this
        assert file_sig == LAS_FILE_SIGNATURE

        header.file_source_id = int.from_bytes(
            stream.read(2), little_endian, signed=False
        )
        header.global_encoding = GlobalEncoding.read_from(stream)

        header.uuid = UUID(bytes_le=stream.read(16))
        header._version = Version(
            int.from_bytes(stream.read(1), little_endian, signed=False),
            int.from_bytes(stream.read(1), little_endian, signed=False),
        )

        header.system_identifier = read_string(stream, SYSTEM_IDENTIFIER_LEN)
        header.generating_software = read_string(stream, GENERATING_SOFTWARE_LEN)

        creation_day_of_year = int.from_bytes(
            stream.read(2), little_endian, signed=False
        )
        creation_year = int.from_bytes(stream.read(2), little_endian, signed=False)
        try:
            header.creation_date = date(creation_year, 1, 1) + timedelta(
                creation_day_of_year - 1
            )
        except ValueError:
            header.creation_date = None

        header_size = int.from_bytes(stream.read(2), little_endian, signed=False)
        header.offset_to_point_data = int.from_bytes(
            stream.read(4), little_endian, signed=False
        )
        number_of_vlrs = int.from_bytes(stream.read(4), little_endian, signed=False)

        point_format_id = int.from_bytes(stream.read(1), little_endian, signed=False)
        point_size = int.from_bytes(stream.read(2), little_endian, signed=False)

        header.point_count = int.from_bytes(stream.read(4), little_endian, signed=False)
        for i in range(5):
            header.number_of_points_by_return[i] = int.from_bytes(
                stream.read(4), little_endian, signed=False
            )

        for i in range(3):
            header.scales[i] = struct.unpack("<d", stream.read(8))[0]
        for i in range(3):
            header.offsets[i] = struct.unpack("<d", stream.read(8))[0]
        for i in range(3):
            header.maxs[i] = struct.unpack("<d", stream.read(8))[0]
            header.mins[i] = struct.unpack("<d", stream.read(8))[0]

        if header.version.minor >= 3:
            header.start_of_waveform_data_packet_record = int.from_bytes(
                stream.read(8), little_endian, signed=False
            )
        if header.version.minor >= 4:
            header.start_of_first_evlr = int.from_bytes(
                stream.read(8), little_endian, signed=False
            )
            header.number_of_evlrs = int.from_bytes(
                stream.read(4), little_endian, signed=False
            )
            header.point_count = int.from_bytes(
                stream.read(8), little_endian, signed=False
            )
            for i in range(15):
                header.number_of_points_by_return[i] = int.from_bytes(
                    stream.read(8), little_endian, signed=False
                )
        if header.version.minor >= 5:
            header.max_gps_time = struct.unpack("<d", stream.read(8))[0]
            header.min_gps_time = struct.unpack("<d", stream.read(8))[0]
            header.gps_time_offset = int.from_bytes(
                stream.read(2), little_endian, signed=False
            )

        current_pos = stream.tell()
        if current_pos < header_size:
            header.extra_header_bytes = stream.read(header_size - current_pos)
        elif current_pos > header_size:
            raise LaspyException("Incoherent header size")

        header._vlrs = VLRList.read_from(stream, num_to_read=number_of_vlrs)

        current_pos = stream.tell()
        if current_pos < header.offset_to_point_data:
            header.extra_vlr_bytes = stream.read(
                header.offset_to_point_data - current_pos
            )
        elif current_pos > header.offset_to_point_data:
            raise LaspyException("Incoherent offset to point data")

        header.are_points_compressed = is_point_format_compressed(point_format_id)
        point_format_id = compressed_id_to_uncompressed(point_format_id)
        point_format = PointFormat(point_format_id)
        try:
            extra_bytes_vlr = typing.cast(
                ExtraBytesVlr, header._vlrs.get("ExtraBytesVlr")[0]
            )
        except IndexError:
            pass
        else:
            if point_size == point_format.size:
                logger.warning(
                    "There is an ExtraByteVlr but the header.point_size matches the "
                    "point size without extra bytes. The extra bytes vlr info will be ignored"
                )
                header._vlrs.extract("ExtraBytesVlr")
            else:
                for extra_dim_info in extra_bytes_vlr.type_of_extra_dims():
                    point_format.add_extra_dimension(extra_dim_info)
        header._point_format = point_format

        if point_size > point_format.size:
            # We have unregistered extra bytes
            num_extra_bytes = point_size - point_format.size
            point_format.dimensions.append(
                dims.DimensionInfo(
                    name="ExtraBytes",
                    kind=dims.DimensionKind.UnsignedInteger,
                    num_bits=8 * num_extra_bytes,
                    num_elements=num_extra_bytes,
                    is_standard=False,
                    description="Un-registered ExtraBytes",
                )
            )
        elif point_size < point_format.size:
            raise LaspyException(
                f"Incoherent point size, "
                f"header says {point_size} point_format created says {point_format.size}"
            )

        if read_evlrs:
            header.read_evlrs(original_stream)
            stream.seek(header.offset_to_point_data)

        return header

    def write_to(
        self,
        stream: BinaryIO,
        ensure_same_size: bool = False,
        encoding_errors: str = "strict",
    ) -> None:
        """
        ensure_same_size: if true this function will raise an internal error
        if the written header would change the offset to point data
        it originally had (meaning the file could become broken),
        Used when rewriting a header to update the file (new point count, mins, maxs, etc)
        """
        if self.point_count > self.max_point_count():
            raise LaspyException(
                f"Version {self.version} cannot save clouds with more than"
                f" {self.max_point_count()} points (current: {self.point_count})"
            )

        little_endian = "little"
        with io.BytesIO() as tmp:
            self._vlrs.write_to(tmp, encoding_errors=encoding_errors)
            vlr_bytes = tmp.getvalue()

        header_size = LAS_HEADERS_SIZE[str(self.version)]
        header_size += len(self.extra_header_bytes)
        new_offset_to_data = header_size + len(vlr_bytes) + len(self.extra_vlr_bytes)

        if ensure_same_size and new_offset_to_data != self.offset_to_point_data:
            raise LaspyException(
                "Internal error, writing header would change original offset to data"
                "and break the file"
            )
        self.offset_to_point_data = new_offset_to_data

        stream.write(LAS_FILE_SIGNATURE)
        stream.write(self.file_source_id.to_bytes(2, little_endian, signed=False))
        self.global_encoding.write_to(stream)
        stream.write(self.uuid.bytes_le)
        stream.write(self.version.major.to_bytes(1, little_endian, signed=False))
        stream.write(self.version.minor.to_bytes(1, little_endian, signed=False))

        was_truncated = write_string(
            stream,
            self.system_identifier,
            SYSTEM_IDENTIFIER_LEN,
            encoding_errors=encoding_errors,
        )
        if was_truncated:
            logger.warning(
                f"system identifier does not fit into the {SYSTEM_IDENTIFIER_LEN} maximum bytes,"
                f" it will be truncated"
            )

        was_truncated = write_string(
            stream,
            self.generating_software,
            GENERATING_SOFTWARE_LEN,
            encoding_errors=encoding_errors,
        )
        if was_truncated:
            logger.warning(
                f"generating software does not fit into the {GENERATING_SOFTWARE_LEN} maximum bytes,"
                f" it will be truncated"
            )

        if self.creation_date is None:
            self.creation_date = date.today()

        stream.write(
            self.creation_date.timetuple().tm_yday.to_bytes(
                2, little_endian, signed=False
            )
        )
        stream.write(self.creation_date.year.to_bytes(2, little_endian, signed=False))

        stream.write(header_size.to_bytes(2, little_endian, signed=False))
        stream.write(self.offset_to_point_data.to_bytes(4, little_endian, signed=False))
        stream.write(len(self._vlrs).to_bytes(4, little_endian, signed=False))

        point_format_id = self.point_format.id
        if self.are_points_compressed:
            point_format_id = uncompressed_id_to_compressed(point_format_id)
        stream.write(point_format_id.to_bytes(1, little_endian, signed=False))
        stream.write(self.point_format.size.to_bytes(2, little_endian, signed=False))

        # Point Count
        if self.version.minor >= 4:
            stream.write(int(0).to_bytes(4, little_endian, signed=False))
            for i in range(5):
                stream.write(int(0).to_bytes(4, little_endian, signed=False))
        else:
            stream.write(self.point_count.to_bytes(4, little_endian, signed=False))
            for i in range(5):
                stream.write(
                    int(self.number_of_points_by_return[i]).to_bytes(
                        4, little_endian, signed=False
                    )
                )

        for i in range(3):
            stream.write(struct.pack("<d", self.scales[i]))
        for i in range(3):
            stream.write(struct.pack("<d", self.offsets[i]))
        for i in range(3):
            stream.write(struct.pack("<d", self.maxs[i]))
            stream.write(struct.pack("<d", self.mins[i]))

        if self.version.minor >= 3:
            stream.write(
                self.start_of_waveform_data_packet_record.to_bytes(
                    8, little_endian, signed=False
                )
            )

        if self.version.minor >= 4:
            stream.write(
                self.start_of_first_evlr.to_bytes(8, little_endian, signed=False)
            )
            stream.write(self.number_of_evlrs.to_bytes(4, little_endian, signed=False))
            stream.write(self.point_count.to_bytes(8, little_endian, signed=False))
            for i in range(15):
                stream.write(
                    int(self.number_of_points_by_return[i]).to_bytes(
                        8, little_endian, signed=False
                    )
                )

        if self.version.minor >= 5:
            stream.write(struct.pack("<d", self.max_gps_time))
            stream.write(struct.pack("<d", self.min_gps_time))
            stream.write(self.gps_time_offset.to_bytes(2, little_endian, signed=False))

        stream.write(self.extra_header_bytes)
        stream.write(vlr_bytes)
        stream.write(self.extra_vlr_bytes)

    def parse_crs(self, prefer_wkt=True) -> Optional["pyproj.CRS"]:
        """
        Method to parse OGC WKT or GeoTiff VLR keys into a pyproj CRS object

        Returns None if no CRS VLR is present, or if the CRS specified
        in the VLRS is not understood.


        Parameters
        ----------
        prefer_wkt: Optional, default True,
            If True the WKT VLR will be preferred in case
            both the WKT and Geotiff VLR are present

        .. warning::
            This requires `pyproj`.

        .. versionadded:: 2.5
            The ``prefer_wkt`` parameters.
        """

        geo_vlr = self._vlrs.get_by_id("LASF_Projection")

        if self.evlrs is not None:
            geo_vlr.extend(self.evlrs.get_by_id("LASF_Projection"))

        parsed_crs = {}
        for rec in geo_vlr:
            if isinstance(rec, (WktCoordinateSystemVlr, GeoKeyDirectoryVlr)):
                crs = rec.parse_crs()
                if crs is not None:
                    parsed_crs[type(rec)] = crs

        # Could not parse anything / there was nothing to parse
        if not parsed_crs:
            return None

        if prefer_wkt:
            preferred, other = WktCoordinateSystemVlr, GeoKeyDirectoryVlr
        else:
            preferred, other = GeoKeyDirectoryVlr, WktCoordinateSystemVlr

        try:
            return parsed_crs[preferred]
        except KeyError:
            return parsed_crs[other]

    def read_evlrs(self, stream):
        """
        Reads EVLRs from the stream and sets them in the
        data property.

        The evlrs are accessed from the `evlrs` property

        Does nothing if either of these is true:
            - The file does not support EVLRS (version < 1.4)
            - The file has no EVLRS
            - The stream does not support seeking

        Leaves/restores the stream position to where it was before the call
        """
        if self.version.minor >= 4:
            if self.number_of_evlrs > 0 and stream.seekable():
                saved_pos = stream.tell()
                stream.seek(self.start_of_first_evlr, io.SEEK_SET)
                self.evlrs = VLRList.read_from(
                    stream, self.number_of_evlrs, extended=True
                )
                stream.seek(saved_pos)
            elif self.number_of_evlrs > 0 and not stream.seekable():
                self.evlrs = None
            else:
                self.evlrs = VLRList()
        else:
            self.evlrs = None

    @staticmethod
    def _prefetch_header_data(source) -> bytes:
        """
        reads (and returns) from the source all the bytes that
        are between the beginning of the file and the start of point data
        (which corresponds to Header + VLRS).

        It is done in two calls to the source's `read` method

        This is done because `LasHeader.read_from`
        does a bunch of read to the source, so we prefer to
        prefetch data in memory in case the original source
        is not buffered (like a http source could be)
        """
        header_bytes = source.read(LAS_HEADERS_SIZE["1.1"])

        file_sig = header_bytes[: len(LAS_FILE_SIGNATURE)]
        if not file_sig:
            raise LaspyException(f"Source is empty")
        if file_sig != LAS_FILE_SIGNATURE:
            raise LaspyException(f'Invalid file signature "{file_sig}"')
        if len(header_bytes) < LAS_HEADERS_SIZE["1.1"]:
            raise LaspyException("File is to small to be a valid LAS")

        offset_to_data = int.from_bytes(
            header_bytes[96 : 96 + 4], byteorder="little", signed=False
        )

        rest = source.read(offset_to_data - len(header_bytes))

        return header_bytes + rest

    def _sync_extra_bytes_vlr(self) -> None:
        try:
            self._vlrs.extract("ExtraBytesVlr")
        except IndexError:
            pass

        extra_dimensions = list(self.point_format.extra_dimensions)
        if not extra_dimensions:
            return

        eb_vlr = ExtraBytesVlr()
        for extra_dimension in extra_dimensions:
            dtype = extra_dimension.dtype
            assert dtype is not None

            if extra_dimension.num_elements > 3 and dtype.base == np.uint8:
                data_type = (0, extra_dimension.num_elements)
            else:
                data_type = extradims.get_id_for_extra_dim_type(dtype)

            eb_struct = ExtraBytesStruct(
                name=extra_dimension.name.encode(),
                description=extra_dimension.description.encode(),
                data_type=data_type,
                scale=extra_dimension.scales,
                offset=extra_dimension.offsets,
                no_data=extra_dimension.no_data,
            )
            eb_vlr.extra_bytes_structs.append(eb_struct)

        self._vlrs.append(eb_vlr)

    # To keep some kind of backward compatibility
    @property
    def major_version(self) -> int:
        return self.version.major

    @property
    def minor_version(self) -> int:
        return self.version.minor

    def __getattr__(self, item):
        try:
            return getattr(self, self._OLD_LASPY_NAMES[item])
        except KeyError:
            raise AttributeError(f"No attribute {item} in LasHeader") from None

    def __setattr__(self, key, value):
        try:
            return setattr(self, self._OLD_LASPY_NAMES[key], value)
        except KeyError:
            super().__setattr__(key, value)

    def __repr__(self) -> str:
        return f"<LasHeader({self.version.major}.{self.version.minor}, {self.point_format})>"


LAS_HEADERS_SIZE = {
    "1.1": 227,
    "1.2": 227,
    "1.3": 235,
    "1.4": 375,
    "1.5": 393,
}



try:
    import rich
    import typer

    HAS_CLI_DEPDS = True
except ModuleNotFoundError:
    HAS_CLI_DEPDS = False


def main():
    if HAS_CLI_DEPDS:
        from .core import app

        app()
    else:
        raise SystemExit(
            """laspy cli needs extra dependencies, install using
`pip install laspy[cli]`
`pip install laspy[cli,lazrs]`
"""
        )


if __name__ == "__main__":
    main()



import logging
import pathlib
import typing
from copy import deepcopy
from typing import BinaryIO, Iterable, List, Optional, Sequence, Union, overload

import numpy as np

from . import errors
from .compression import LazBackend
from .header import LasHeader
from .laswriter import LasWriter
from .point import ExtraBytesParams, PointFormat, dims, record
from .point.dims import OLD_LASPY_NAMES, ScaledArrayView, SubFieldView
from .point.record import DimensionNameValidity
from .vlrs.vlrlist import VLRList

logger = logging.getLogger(__name__)


class LasData:
    """Class synchronizing all the moving parts of LAS files.

    It connects the point record, header, vlrs together.

    To access points dimensions using this class you have two possibilities

    .. code:: python

        las = laspy.read('some_file.las')
        las.classification
        # or
        las['classification']
    """

    def __init__(
        self,
        header: LasHeader,
        points: Optional[
            Union[record.PackedPointRecord, record.ScaleAwarePointRecord]
        ] = None,
    ) -> None:
        if points is None:
            points = record.ScaleAwarePointRecord.zeros(
                header.point_count, header=header
            )
        if points.point_format != header.point_format:
            raise errors.LaspyException("Incompatible Point Formats")
        if isinstance(points, record.PackedPointRecord):
            points = record.ScaleAwarePointRecord(
                points.array,
                header.point_format,
                scales=header.scales,
                offsets=header.offsets,
            )
        else:
            assert np.all(header.scales, points.scales)
            assert np.all(header.offsets, points.offsets)
        self.__dict__["_points"] = points
        self.points: record.ScaleAwarePointRecord
        self.header: LasHeader = header

    @property
    def point_format(self) -> PointFormat:
        """Shortcut to get the point format"""
        return self.points.point_format

    @property
    def xyz(self) -> np.ndarray:
        """Returns a **new** 2D numpy array with the x,y,z coordinates

        >>> import laspy
        >>> las = laspy.read("tests/data/simple.las")
        >>> xyz = las.xyz
        >>> xyz.ndim
        2
        >>> xyz.shape
        (1065, 3)
        >>> bool(np.all(xyz[..., 0] == las.x))
        True
        """
        return np.vstack((self.x, self.y, self.z)).transpose()

    @xyz.setter
    def xyz(self, value) -> None:
        self.points[("x", "y", "z")] = value

    @property
    def points(self) -> record.PackedPointRecord:
        """Returns the point record"""
        return self._points

    @points.setter
    def points(self, new_points: record.PackedPointRecord) -> None:
        if new_points.point_format != self.point_format:
            raise errors.IncompatibleDataFormat(
                "Cannot set points with a different point format, convert first"
            )
        self._points = new_points
        self.update_header()
        # make sure both point format point to the same object
        self._points.point_format = self.header.point_format

    @property
    def vlrs(self) -> VLRList:
        return self.header.vlrs

    @vlrs.setter
    def vlrs(self, vlrs) -> None:
        self.header.vlrs = vlrs

    @property
    def evlrs(self) -> Optional[VLRList]:
        return self.header.evlrs

    @evlrs.setter
    def evlrs(self, evlrs: VLRList) -> None:
        self.header.evlrs = evlrs

    def add_extra_dim(self, params: ExtraBytesParams) -> None:
        """Adds a new extra dimension to the point record

        .. note::

            If you plan on adding multiple extra dimensions,
            prefer :meth:`.add_extra_dims` as it will
            save re-allocations and data copy

        Parameters
        ----------
        params : ExtraBytesParams
            parameters of the new extra dimension to add

        """
        self.add_extra_dims([params])

    def add_extra_dims(self, params: List[ExtraBytesParams]) -> None:
        """Add multiple extra dimensions at once

        Parameters
        ----------

        params: list of parameters of the new extra dimensions to add
        """
        self.header.add_extra_dims(params)
        new_point_record = record.ScaleAwarePointRecord.zeros(
            len(self.points), header=self.header
        )
        new_point_record.copy_fields_from(self.points)
        self.points = new_point_record

    def remove_extra_dims(self, names: Iterable[str]) -> None:
        """Remove multiple extra dimensions from this object

        Parameters
        ----------

        names: iterable,
            names of the extra dimensions to be removed


        Raises
        ------

        LaspyException: if you try to remove an extra dimension that do not exist.

        """
        extra_dimension_names = list(self.point_format.extra_dimension_names)
        not_extra_dimension = [
            name for name in names if name not in extra_dimension_names
        ]
        if not_extra_dimension:
            raise errors.LaspyException(
                f"'{not_extra_dimension}' are not extra dimensions and cannot be removed"
            )

        self.header.remove_extra_dims(names)
        new_point_record = record.ScaleAwarePointRecord.zeros(
            len(self.points), header=self.header
        )
        new_point_record.copy_fields_from(self.points)
        self.points = new_point_record

    def remove_extra_dim(self, name: str) -> None:
        """Remove an extra dimensions from this object

        .. note::

             If you plan on removing multiple extra dimensions,
             prefer :meth:`.remove_extra_dims` as it will
             save re-allocations and data copy

        Parameters
        ----------

        name: str,
            name of the extra dimension to be removed


        Raises
        ------

        LaspyException: if you try to remove an extra dimension that do not exist.

        """
        self.remove_extra_dims([name])

    def update_header(self) -> None:
        """Update the information stored in the header
        to be in sync with the actual data.

        This method is called automatically when you save a file using
        :meth:`laspy.lasdatas.base.LasBase.write`
        """
        self.header.update(self.points)
        self.header.point_format_id = self.points.point_format.id
        self.header.point_data_record_length = self.points.point_size

        if self.header.version.minor >= 4:
            if self.evlrs is not None:
                self.header.number_of_evlrs = len(self.evlrs)
            self.header.start_of_waveform_data_packet_record = 0
            # TODO
            # if len(self.vlrs.get("WktCoordinateSystemVlr")) == 1:
            #     self.header.global_encoding.wkt = 1
        else:
            self.header.number_of_evlrs = 0

    @overload
    def write(
        self,
        destination: str,
        laz_backend: Optional[Union[LazBackend, Sequence[LazBackend]]] = ...,
    ) -> None: ...

    @overload
    def write(
        self,
        destination: BinaryIO,
        do_compress: Optional[bool] = ...,
        laz_backend: Optional[Union[LazBackend, Sequence[LazBackend]]] = ...,
    ) -> None: ...

    def write(self, destination, do_compress=None, laz_backend=None):
        """Writes to a stream or file

        .. note::

            When destination is a string, it will be interpreted as the path were the file should be written to,
            and whether the file will be compressed depends on the extension used (case insensitive):

                - .laz -> compressed
                - .las -> uncompressed

            And the do_compress option will be ignored


        Parameters
        ----------
        destination: str or file object
            filename or stream to write to
        do_compress: bool, optional
            Flags to indicate if you want to compress the data
        laz_backend: optional, the laz backend to use
            By default, laspy detect available backends
        """
        if isinstance(destination, (str, pathlib.Path)):
            do_compress = pathlib.Path(destination).suffix.lower() == ".laz"

            with open(destination, mode="wb+") as out:
                self._write_to(out, do_compress=do_compress, laz_backend=laz_backend)
        else:
            self._write_to(
                destination, do_compress=do_compress, laz_backend=laz_backend
            )

    def _write_to(
        self,
        out_stream: BinaryIO,
        do_compress: Optional[bool] = None,
        laz_backend: Optional[Union[LazBackend, Sequence[LazBackend]]] = None,
    ) -> None:
        with LasWriter(
            out_stream,
            self.header,
            do_compress=do_compress,
            closefd=False,
            laz_backend=laz_backend,
        ) as writer:
            writer.write_points(self.points)
            if self.header.version.minor >= 4 and self.evlrs is not None:
                writer.write_evlrs(self.evlrs)

    def change_scaling(self, scales=None, offsets=None) -> None:
        """This changes the scales and/or offset used for the x,y,z
        dimensions.

        It recomputes the internal, non-scaled X,Y,Z dimensions
        to match the new scales and offsets.

        It also updates the header with the new values of scales and offsets.

        Parameters
        ----------
        scales: optional
                New scales to be used. If not provided, the scales won't change.
        offsets: optional
                 New offsets to be used. If not provided, the offsets won't change.

        Example
        -------

        >>> import laspy
        >>> header = laspy.LasHeader()
        >>> header.scales = np.array([0.1, 0.1, 0.1])
        >>> header.offsets = np.array([0, 0 ,0])

        >>> las = laspy.LasData(header=header)
        >>> las.x = [10.0]
        >>> las.y = [20.0]
        >>> las.z = [30.0]

        >>> # X = (x - x_offset) / x_scale
        >>> assert np.all(las.xyz == [[10.0, 20., 30]])
        >>> assert np.all(las.X == [100])
        >>> assert np.all(las.Y == [200])
        >>> assert np.all(las.Z == [300])

        We change the scales (only changing x_scale here)
        but not the offsets.

        The xyz coordinates (double) are the same (minus possible rounding with actual coordinates)
        However the integer coordinates changed

        >>> las.change_scaling(scales=[0.01, 0.1, 0.1])
        >>> assert np.all(las.xyz == [[10.0, 20., 30]])
        >>> assert np.all(las.X == [1000])
        >>> assert np.all(las.Y == [200])
        >>> assert np.all(las.Z == [300])

        Same idea if we change the offsets, the xyz do not change
        but XYZ does

        >>> las.change_scaling(offsets=[0, 10, 15])
        >>> assert np.all(las.xyz == [[10.0, 20., 30]])
        >>> assert np.all(las.X == [1000])
        >>> assert np.all(las.Y == [100])
        >>> assert np.all(las.Z == [150])
        """
        self.points.change_scaling(scales, offsets)

        if scales is not None:
            self.header.scales = scales
        if offsets is not None:
            self.header.offsets = offsets

    def __getattr__(self, item):
        """Automatically called by Python when the attribute
        named 'item' is no found. We use this function to forward the call the
        point record. This is the mechanism used to allow the users to access
        the points dimensions directly through a LasData.

        Parameters
        ----------
        item: str
            name of the attribute, should be a dimension name

        Returns
        -------
        The requested dimension if it exists

        """
        try:
            return self.points[item]
        except ValueError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{item}'"
            ) from None

    def __setattr__(self, key, value):
        """This is called on every access to an attribute of the instance.
        Again we use this to forward the call the the points record

        But this time checking if the key is actually a dimension name
        so that an error is raised if the user tries to set a valid
        LAS dimension even if it is not present in the field.
        eg: user tries to set the red field of a file with point format 0:
        an error is raised
        """
        if key in ("x", "y", "z"):
            # It is possible that user created a `LasData` object
            # via `laspy.create`, and changed the headers offsets and scales
            # values afterwards. So we need to sync the points's record.
            self.points.offsets = self.header.offsets
            self.points.scales = self.header.scales
            self.points[key] = value
            return

        name_validity = self.points.validate_dimension_name(key)
        if name_validity == DimensionNameValidity.Valid:
            self[key] = value
        elif name_validity == DimensionNameValidity.Unsupported:
            raise ValueError(
                f"Point format {self.point_format} does not support {key} dimension"
            )
        else:
            super().__setattr__(key, value)

    @typing.overload
    def __getitem__(
        self, item: Union[str, List[str]]
    ) -> Union[np.ndarray, ScaledArrayView, SubFieldView]: ...

    @typing.overload
    def __getitem__(
        self, item: Union[int, typing.Iterable[int], slice]
    ) -> "LasData": ...

    def __getitem__(self, item):
        try:
            item_is_list_of_str = all(isinstance(el, str) for el in iter(item))
        except TypeError:
            item_is_list_of_str = False

        if isinstance(item, str) or item_is_list_of_str:
            return self.points[item]
        else:
            las = LasData(deepcopy(self.header), points=self.points[item])
            las.update_header()
            return las

    def __setitem__(self, key, value):
        self.points[key] = value

    def __len__(self):
        return len(self.points)

    def __repr__(self) -> str:
        return "<LasData({}.{}, point fmt: {}, {} points, {} vlrs)>".format(
            self.header.version.major,
            self.header.version.minor,
            self.points.point_format,
            len(self.points),
            len(self.vlrs),
        )



from typing import Any, BinaryIO, List, Optional, Tuple, Union, cast

import numpy as np

from .._pointappender import IPointAppender
from .._pointreader import IPointReader
from .._pointwriter import IPointWriter
from ..header import LasHeader
from ..point.format import PointFormat
from ..point.record import PackedPointRecord
from ..vlrs.known import LasZipVlr
from .lazbackend import ILazBackend
from .selection import DecompressionSelection

try:
    import lazrs
except ModuleNotFoundError:
    lazrs = None


class LazrsBackend(ILazBackend):
    def __init__(
        self,
        parallel: bool = True,
    ):
        self._parallel = parallel

    def is_available(self) -> bool:
        return lazrs is not None

    @property
    def supports_append(self) -> bool:
        return True

    def create_appender(self, dest: BinaryIO, header: LasHeader) -> IPointAppender:
        return LazrsAppender(dest, header, parallel=self._parallel)

    def create_reader(
        self,
        source: Any,
        header: LasHeader,
        decompression_selection: Optional[DecompressionSelection] = None,
    ) -> IPointReader:
        if decompression_selection is None:
            decompression_selection = DecompressionSelection.all()
        laszip_vlr: LasZipVlr = header.vlrs[header.vlrs.index("LasZipVlr")]
        return LazrsPointReader(
            source,
            laszip_vlr,
            parallel=self._parallel,
            decompression_selection=decompression_selection,
        )

    def create_writer(
        self,
        dest: Any,
        header: "LasHeader",
    ) -> IPointWriter:
        return LazrsPointWriter(dest, header.point_format, parallel=self._parallel)


class LazrsPointReader(IPointReader):
    """Implementation for the laz-rs backend, supports single-threaded decompression
    as well as multi-threaded decompression
    """

    def __init__(
        self,
        source,
        laszip_vlr: LasZipVlr,
        parallel: bool,
        decompression_selection: DecompressionSelection,
    ) -> None:
        self._source = source
        self.vlr = lazrs.LazVlr(laszip_vlr.record_data)
        selection = decompression_selection.to_lazrs()
        if parallel:
            self.decompressor = lazrs.ParLasZipDecompressor(
                source, laszip_vlr.record_data, selection
            )
        else:
            self.decompressor = lazrs.LasZipDecompressor(
                source, laszip_vlr.record_data, selection
            )

    @property
    def source(self):
        return self._source

    def read_n_points(self, n: int) -> bytearray:
        point_bytes = bytearray(n * self.vlr.item_size())
        self.decompressor.decompress_many(point_bytes)
        return point_bytes

    def seek(self, point_index: int) -> None:
        self.decompressor.seek(point_index)

    def close(self) -> None:
        self.source.close()

    def read_chunk_table_only(self) -> List[Tuple[int, int]]:
        """
        This function requires the source to be at the start of the chunk table
        """
        assert isinstance(self.decompressor, lazrs.LasZipDecompressor)
        return self.decompressor.read_chunk_table_only()

    def read_raw_bytes(self, n: int) -> bytes:
        """
        reads and returns exactly `n` bytes from the source used by
        this point reader.
        """
        b = bytearray(n)
        self.decompressor.read_raw_bytes_into(b)
        return bytes(b)


class LazrsPointWriter(IPointWriter):
    """
    Compressed point writer using lazrs backend
    """

    def __init__(
        self, dest: BinaryIO, point_format: PointFormat, parallel: bool
    ) -> None:
        self.dest = dest
        self.vlr = lazrs.LazVlr.new_for_compression(
            point_format.id, point_format.num_extra_bytes
        )
        self.parallel = parallel
        self.compressor: Optional[
            Union[lazrs.ParLasZipCompressor, lazrs.LasZipCompressor]
        ] = None

    def write_initial_header_and_vlrs(
        self, header: LasHeader, encoding_errors: str
    ) -> None:
        laszip_vlr = LasZipVlr(self.vlr.record_data())
        header.vlrs.append(laszip_vlr)
        super().write_initial_header_and_vlrs(header, encoding_errors)
        # We have to initialize our compressor here
        # because on init, it writes the offset to chunk table
        # so the header and vlrs have to be written
        if self.parallel:
            self.compressor = lazrs.ParLasZipCompressor(self.dest, self.vlr)
        else:
            self.compressor = lazrs.LasZipCompressor(self.dest, self.vlr)

    @property
    def destination(self) -> BinaryIO:
        return self.dest

    def write_points(self, points: PackedPointRecord) -> None:
        assert (
            self.compressor is not None
        ), "Trying to write points without having written header"
        points_bytes = np.frombuffer(points.array, np.uint8)
        self.compressor.compress_many(points_bytes)

    def done(self) -> None:
        if self.compressor is not None:
            self.compressor.done()


class LazrsAppender(IPointAppender):
    """Appending in LAZ file
    works by seeking to start of the last chunk
    of compressed points, decompress it while keeping the points in
    memory.

    Then seek back to the start of the last chunk, and recompress
    the points we just read, so that we have a compressor in the proper state
    ready to compress new points.
    """

    def __init__(self, dest: BinaryIO, header: LasHeader, parallel: bool) -> None:
        self.offset_to_point_data = header.offset_to_point_data
        laszip_vlr = cast(LasZipVlr, header.vlrs.get("LasZipVlr")[0])
        if parallel:
            try:
                self.appender = lazrs.ParLasZipAppender(
                    dest, laszip_vlr.record_data, header.point_count
                )
            except TypeError:
                # This is a fallback for when lazrs <= 0.7.0
                self.appender = lazrs.ParLasZipAppender(dest, laszip_vlr.record_data)
        else:
            try:
                self.appender = lazrs.LasZipAppender(
                    dest, laszip_vlr.record_data, header.point_count
                )
            except TypeError:
                # This is a fallback for when lazrs <= 0.7.0
                self.appender = lazrs.LasZipAppender(dest, laszip_vlr.record_data)

    def append_points(self, points: PackedPointRecord) -> None:
        points_bytes = np.frombuffer(points.array, np.uint8)
        self.appender.compress_many(points_bytes)

    def done(self) -> None:
        self.appender.done()



"""The definition of the VLR Header, VLR, the KnownVLRs
are in this module.

A KnownVLR is a VLR for which we know how to parse its record_data
"""

import abc
import ctypes
import logging
import struct
from copy import copy
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

import numpy as np

from ..extradims import get_dtype_for_extra_dim
from ..point.dims import ScaledArrayView
from ..point.format import ExtraBytesParams
from ..point.record import PackedPointRecord
from ..utils import encode_to_null_terminated
from .vlr import VLR, BaseVLR

abstractmethod = abc.abstractmethod

logger = logging.getLogger(__name__)

NULL_BYTE = b"\0"

GeoKeyDirectoryType = TypeVar("GeoKeyDirectoryType", bound="GeoKeyDirectoryVlr")
GeoAsciiParamsType = TypeVar("GeoAsciiParamsType", bound="GeoAsciiParamsVlr")


class IKnownVLR(abc.ABC):
    """Interface that any KnownVLR must implement.
    A KnownVLR is a VLR for which we know how to parse its record_data

    Implementing this interfaces allows to automatically call the
    right parser for the right VLR when reading them.
    """

    @staticmethod
    @abstractmethod
    def official_user_id() -> str:
        """Shall return the official user_id as described in the documentation"""
        pass

    @staticmethod
    @abstractmethod
    def official_record_ids() -> Tuple[int, ...]:
        """Shall return the official record_id for the VLR

        .. note::

            Even if the VLR has one record_id, the return type must be a tuple

        Returns
        -------
        tuple of int
            The record_ids this VLR type can have
        """
        pass

    @abstractmethod
    def record_data_bytes(self) -> bytes:
        """Shall return the bytes corresponding to the record_data part of the VLR
        as they should be written in the file.

        Returns
        -------
        bytes
            The bytes of the vlr's record_data

        """
        pass

    @abstractmethod
    def parse_record_data(self, record_data: bytes) -> None:
        """Shall parse the given record_data into a user-friendlier structure

        Parameters
        ----------
        record_data: bytes
            The record_data bytes read from the file

        """
        pass


class BaseKnownVLR(BaseVLR, IKnownVLR, abc.ABC):
    """Base Class to factorize common code between the different type of Known VLRs"""

    def __init__(self, record_id=None, description=""):
        super().__init__(
            self.official_user_id(),
            self.official_record_ids()[0] if record_id is None else record_id,
            description,
        )

    @classmethod
    def from_raw(cls, raw: VLR):
        know_vlr = cls()
        know_vlr._description = raw.description
        know_vlr.parse_record_data(raw.record_data)
        return know_vlr


class ClassificationLookupVlr(BaseKnownVLR):
    """This vlr maps class numbers to short descriptions / names

    >>> lookup = ClassificationLookupVlr()
    >>> lookup[0] = "never_classified"
    >>> lookup[2] = "ground"
    >>> lookup[0]
    'never_classified'
    """

    _lookup_struct = struct.Struct("<B15s")

    def __init__(self):
        super().__init__(description="Classification Lookup")
        self.lookups: Dict[int, str] = {}

    def parse_record_data(self, record_data: bytes) -> None:
        for class_id, desc in struct.iter_unpack("<B15s", record_data):
            # index using desc[i:i+1], because desc[i] gives an int, and we want a byte
            description = b"".join(
                desc[i : i + 1]
                for i in range(len(desc))
                if desc[i : i + 1].isalnum() or desc[i : i + 1] == b" "
            ).decode()
            self.lookups[class_id] = description

    def record_data_bytes(self) -> bytes:
        def lookup_converter(lookup_dict):
            for class_id, description in lookup_dict.items():
                description_bytes = description.encode("ascii")
                if len(description_bytes) > 15:
                    raise ValueError(
                        "decription ({}) is to long ({} bytes), it must not exceed 15 bytes when encoded".format(
                            description, len(description_bytes)
                        )
                    )
                yield class_id, description_bytes

        return b"".join(
            self._lookup_struct.pack(class_id, desc)
            for class_id, desc in lookup_converter(self.lookups)
        )

    def __getitem__(self, class_id: int) -> str:
        return self.lookups[class_id]

    def __setitem__(self, class_id: int, description: str):
        if class_id not in range(256):
            raise ValueError("Class id {} is not in range [0, 255]".format(class_id))

        self.lookups[class_id] = description

    @staticmethod
    def official_user_id() -> str:
        return "LASF_Spec"

    @staticmethod
    def official_record_ids() -> Tuple[int, ...]:
        return (0,)


class LasZipVlr(BaseKnownVLR):
    """Contains the information needed by laszip (or any other laz backend)
    to compress the point records.
    """

    def __init__(self, data: bytes) -> None:
        super().__init__(description="http://laszip.org")
        self.record_data = data

    def parse_record_data(self, record_data: bytes) -> None:
        # Only laz backends know how to parse this
        pass

    def record_data_bytes(self) -> bytes:
        return self.record_data

    @staticmethod
    def official_user_id() -> str:
        return "laszip encoded"

    @staticmethod
    def official_record_ids() -> Tuple[int, ...]:
        return (22204,)

    @classmethod
    def from_raw(cls, raw_vlr):
        return cls(raw_vlr.record_data)


class ExtraBytesStruct(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("reserved", ctypes.c_uint8 * 2),
        ("data_type", ctypes.c_uint8),
        ("options", ctypes.c_uint8),
        ("name", ctypes.c_char * 32),
        ("unused", ctypes.c_uint8 * 4),
        ("_no_data", (ctypes.c_byte * 8) * 3),
        ("_min", (ctypes.c_byte * 8) * 3),
        ("_max", (ctypes.c_byte * 8) * 3),
        ("_scale", ctypes.c_double * 3),
        ("_offset", ctypes.c_double * 3),
        ("description", ctypes.c_char * 32),
    ]

    _uint64t_struct = struct.Struct("<Q")
    _int64t_struct = struct.Struct("<q")
    _double_struct = struct.Struct("<d")

    NO_DATA_BIT_MASK = 0b000_0001
    MIN_BIT_MASK = 0b0000_0010
    MAX_BIT_MASK = 0b0000_0100
    SCALE_BIT_MASK = 0b000_1000
    OFFSET_BIT_MASK = 0b0001_0000

    def __init__(
        self,
        name: bytes,
        data_type: Union[int, Tuple[int, int]],
        description: bytes = b"",
        scale: Optional[np.ndarray] = None,
        offset: Optional[np.ndarray] = None,
        no_data: Optional[np.ndarray] = None,
    ) -> None:

        if isinstance(data_type, Tuple):
            options = data_type[1]
            data_type = data_type[0]
        else:
            options = 0

        super().__init__(
            name=name, description=description, data_type=data_type, options=options
        )

        if self.data_type != 0:
            self.scale = scale
            self.offset = offset
            self.no_data = no_data

            self.options |= self.MIN_BIT_MASK
            self.options |= self.MAX_BIT_MASK

            self.partial_reset()

    def _long_type(self):
        data_type = ((self.data_type - 1) % 10) + 1
        if data_type in [2, 4, 6, 8]:
            long_type = np.int64
        elif data_type in [9, 10]:
            long_type = np.float64
        elif data_type in [1, 3, 5, 7]:
            long_type = np.uint64
        else:
            raise NotImplementedError

        return long_type

    def _parse_special_property(self, name) -> np.ndarray:
        dtype = self.dtype().base
        long_type = self._long_type()
        return np.frombuffer(getattr(self, name), dtype=long_type)[
            : self.num_elements()
        ].astype(dtype)

    @property
    def no_data(self):
        if self.options & self.NO_DATA_BIT_MASK != 0:
            return self._parse_special_property("_no_data")
        return None

    @no_data.setter
    def no_data(self, value):
        if value is None:
            self.options &= ~self.NO_DATA_BIT_MASK
        else:
            dtype = self._long_type()
            num_elements = self.num_elements()
            ptrs = [
                np.array([v])
                .astype(dtype)
                .ctypes.data_as(ctypes.POINTER((ctypes.c_byte * 8)))[0]
                for v in value[:num_elements]
            ]
            self._no_data[:num_elements] = ptrs
            self.options |= self.NO_DATA_BIT_MASK

    @property
    def min(self):
        if not self.min_is_relevant():
            return None

        min = self._parse_special_property("_min")

        scale = self.scale
        if scale is not None:
            min = min * scale

        offset = self.offset
        if offset is not None:
            min = min + offset

        return min

    @property
    def max(self):
        if not self.max_is_relevant():
            return None

        max = self._parse_special_property("_max")

        scale = self.scale
        if scale is not None:
            max = max * scale

        offset = self.offset
        if offset is not None:
            max = max + offset

        return max

    @property
    def offset(self) -> Optional[Any]:
        if self.options & self.OFFSET_BIT_MASK != 0:
            return self._offset[: self.num_elements()]
        return None

    @offset.setter
    def offset(self, value):
        if value is None:
            self.options &= ~self.OFFSET_BIT_MASK
        else:
            num_elements = self.num_elements()
            self._offset[:num_elements] = value[:num_elements]
            self.options |= self.OFFSET_BIT_MASK

    @property
    def scale(self):
        if self.options & self.SCALE_BIT_MASK != 0:
            return self._scale[: self.num_elements()]
        return None

    @scale.setter
    def scale(self, value):
        if value is None:
            self.options &= ~self.SCALE_BIT_MASK
        else:
            num_elements = self.num_elements()
            self._scale[:num_elements] = value[:num_elements]
            self.options |= self.SCALE_BIT_MASK

    def format_name(self):
        return self.name.rstrip(NULL_BYTE).decode()

    def dtype(self) -> np.dtype:
        if self.data_type == 0:
            if self.options == 1:
                # numpy says doing '1u1' is deprecated
                return np.dtype("u1")
            return np.dtype(f"{self.options}u1")
        return get_dtype_for_extra_dim(self.data_type)

    def num_elements(self) -> int:
        if self.data_type == 0:
            return self.options
        elif self.data_type <= 10:
            return 1
        elif self.data_type <= 20:
            return 2
        else:
            return 3

    def min_is_relevant(self):
        return self.options & self.MIN_BIT_MASK != 0

    def max_is_relevant(self):
        return self.options & self.MAX_BIT_MASK != 0

    def _raw_min(self):
        if self.min_is_relevant():
            return np.frombuffer(self._min, dtype=self._long_type())[
                : self.num_elements()
            ]
        return None

    def _raw_max(self):
        if self.max_is_relevant():
            return np.frombuffer(self._max, dtype=self._long_type())[
                : self.num_elements()
            ]
        return None

    def grow(self, points: PackedPointRecord):
        if self.data_type == 0:
            # The data type is not specified (treated as raw bytes)
            # So we don't try to track min / max
            return

        if not self.min_is_relevant() and not self.max_is_relevant():
            return

        try:
            pts = points[self.format_name()]
        except ValueError:
            # The points did not contain the field
            return

        num_elements = self.num_elements()
        long_type = self._long_type()
        no_data = self.no_data

        local_min = np.zeros(num_elements, dtype=long_type)
        local_max = np.zeros(num_elements, dtype=long_type)

        for i in range(num_elements):
            if no_data is not None:
                valid_indices = pts[..., i] != no_data[i]
                if valid_indices.ndim == 0:
                    return
                sub_pts = pts[valid_indices, i]
            else:
                sub_pts = pts[..., i]

            if self.min_is_relevant():
                if isinstance(sub_pts, ScaledArrayView):
                    local_min[i] = sub_pts.array.min()
                else:
                    local_min[i] = sub_pts.min()

            if self.max_is_relevant():
                if isinstance(sub_pts, ScaledArrayView):
                    local_max[i] = sub_pts.array.max()
                else:
                    local_max[i] = sub_pts.max()

        if self.min_is_relevant():
            raw_min = self._raw_min()
            v = np.min([raw_min, local_min], axis=0)
            raw_min[:] = v.astype(long_type)[:]

        if self.max_is_relevant():
            raw_max = self._raw_max()
            v = np.max([raw_max, local_max], axis=0)
            raw_max[:] = v.astype(long_type)[:]

    def partial_reset(self):
        long_type = self._long_type()

        if long_type == np.float64 or long_type == np.float32:
            info = np.finfo(long_type)
        else:
            info = np.iinfo(long_type)

        num_elements = self.num_elements()
        np.frombuffer(self._min, dtype=long_type)[:num_elements] = info.max
        np.frombuffer(self._max, dtype=long_type)[:num_elements] = info.min

    @staticmethod
    def size():
        return ctypes.sizeof(ExtraBytesStruct)

    def __repr__(self):
        return "<ExtraBytesStruct({}, {}, {})>".format(
            self.format_name(), self.data_type, self.description
        )


class ExtraBytesVlr(BaseKnownVLR):
    def __init__(self):
        super().__init__(description="Extra Bytes Record")
        self.extra_bytes_structs: List[ExtraBytesStruct] = []

    def parse_record_data(self, data):
        if (len(data) % ExtraBytesStruct.size()) != 0:
            raise ValueError(
                "Data length of ExtraBytes vlr must be a multiple of {}".format(
                    ExtraBytesStruct.size()
                )
            )
        num_extra_bytes_structs = len(data) // ExtraBytesStruct.size()
        self.extra_bytes_structs = [None] * num_extra_bytes_structs
        for i in range(num_extra_bytes_structs):
            self.extra_bytes_structs[i] = ExtraBytesStruct.from_buffer_copy(
                data[ExtraBytesStruct.size() * i : ExtraBytesStruct.size() * (i + 1)]
            )

    def record_data_bytes(self):
        return b"".join(
            bytes(extra_struct) for extra_struct in self.extra_bytes_structs
        )

    def type_of_extra_dims(self) -> List[ExtraBytesParams]:
        dim_info_list: List[ExtraBytesParams] = []
        for eb_struct in self.extra_bytes_structs:
            num_elements = eb_struct.num_elements()

            scales = eb_struct.scale
            offsets = eb_struct.offset

            if scales is not None or offsets is not None:
                # If one of scales or offsets is defined,
                # we expect the other to be as well
                # so set default scales or offsets
                if offsets is None:
                    offsets = np.zeros(num_elements, np.float64)
                else:
                    offsets = np.array(offsets[:num_elements])

                if scales is None:
                    scales = np.ones(num_elements, np.float64)
                else:
                    scales = np.array(scales[:num_elements])

            dim_info_list.append(
                ExtraBytesParams(
                    eb_struct.format_name(),
                    eb_struct.dtype(),
                    description=eb_struct.description.rstrip(NULL_BYTE).decode(),
                    scales=scales,
                    offsets=offsets,
                )
            )
        return dim_info_list

    def grow(self, points: PackedPointRecord):
        for eb_struct in self.extra_bytes_structs:
            eb_struct.grow(points)

    def partial_reset(self):
        for eb_struct in self.extra_bytes_structs:
            eb_struct.partial_reset()

    def __repr__(self):
        return "<ExtraBytesVlr(extra bytes structs: {})>".format(
            len(self.extra_bytes_structs)
        )

    @staticmethod
    def official_user_id():
        return "LASF_Spec"

    @staticmethod
    def official_record_ids():
        return (4,)


class WaveformPacketStruct(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("bits_per_sample", ctypes.c_ubyte),
        ("waveform_compression_type", ctypes.c_ubyte),
        ("number_of_samples", ctypes.c_uint32),
        ("temporal_sample_spacing", ctypes.c_uint32),
        ("digitizer_gain", ctypes.c_double),
        ("digitizer_offset", ctypes.c_double),
    ]

    @staticmethod
    def size():
        return ctypes.sizeof(WaveformPacketStruct)


class WaveformPacketVlr(BaseKnownVLR):
    def __init__(self, record_id, description=""):
        super().__init__(record_id=record_id, description=description)
        self.parsed_record = None

    def parse_record_data(self, record_data):
        self.parsed_record = WaveformPacketStruct.from_buffer_copy(record_data)

    def record_data_bytes(self):
        return bytes(self.parsed_record)

    @staticmethod
    def official_record_ids():
        return range(100, 356)

    @staticmethod
    def official_user_id():
        return "LASF_Spec"

    @classmethod
    def from_raw(cls, raw_vlr):
        vlr = cls(raw_vlr.record_id, description=raw_vlr.description)
        vlr._description = raw_vlr.description
        vlr.parse_record_data(raw_vlr.record_data)
        return vlr


class GeoKeyEntryStruct(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        # Id of the key
        #
        # Ids are broken down in sub domains:
        # [    0,  1023]       Reserved
        # [ 1024,  2047]       GeoTIFF Configuration Keys
        # [ 2048,  3071]       Geographic/Geocentric CS Parameter Keys
        # [ 3072,  4095]       Projected CS Parameter Keys
        # [ 4096,  5119]       Vertical CS Parameter Keys
        # [ 5120, 32767]       Reserved
        # [32768, 65535]       Private use
        ("id", ctypes.c_uint16),
        # Where to find the data for the key:
        # 0 => The _actual_ value is stored directly in the "value_offset" member
        # Otherwise, the tiff tag location is the record_id of the VLR in which the value is stored.
        # In the case of LAS files the 2 possible values are `34736`, `34737`.
        ("tiff_tag_location", ctypes.c_uint16),
        # Number of values in the key.
        # Implied to be `1` if `tiff_tag_location` is 0
        ("count", ctypes.c_uint16),
        # Depending on `tiff_tag_location`, this contains either
        # the value itself _or_ the offset in the record_data of the containing VLR
        ("value_offset", ctypes.c_uint16),
    ]

    @staticmethod
    def size():
        return ctypes.sizeof(GeoKeysHeaderStructs)

    def __repr__(self):
        return "<GeoKey(Id: {}, Location: {}, count: {}, offset: {})>".format(
            self.id, self.tiff_tag_location, self.count, self.value_offset
        )


class GeoKeysHeaderStructs(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("key_directory_version", ctypes.c_uint16),
        ("key_revision", ctypes.c_uint16),
        ("minor_revision", ctypes.c_uint16),
        ("number_of_keys", ctypes.c_uint16),
    ]

    def __init__(self):
        super().__init__(
            key_directory_version=1, key_revision=1, minor_revision=0, number_of_keys=0
        )

    @staticmethod
    def size():
        return ctypes.sizeof(GeoKeysHeaderStructs)

    def __repr__(self):
        return "<GeoKeysHeader(vers: {}, rev:{}, minor: {}, num_keys: {})>".format(
            self.key_directory_version,
            self.key_revision,
            self.minor_revision,
            self.number_of_keys,
        )


class GeoKeyDirectoryVlr(BaseKnownVLR):
    def __init__(self):
        super().__init__(description="GeoTIFF GeoKeyDirectoryTag")
        self.geo_keys_header = GeoKeysHeaderStructs()
        self.geo_keys = [GeoKeyEntryStruct()]

    def parse_record_data(self, record_data):
        record_data = bytearray(record_data)
        header_data = record_data[: ctypes.sizeof(GeoKeysHeaderStructs)]
        self.geo_keys_header = GeoKeysHeaderStructs.from_buffer(header_data)
        self.geo_keys = []
        keys_data = record_data[GeoKeysHeaderStructs.size() :]
        num_keys = (
            len(record_data[GeoKeysHeaderStructs.size() :]) // GeoKeyEntryStruct.size()
        )
        if num_keys != self.geo_keys_header.number_of_keys:
            self.geo_keys_header.number_of_keys = num_keys

        for i in range(self.geo_keys_header.number_of_keys):
            data = keys_data[
                (i * GeoKeyEntryStruct.size()) : (i + 1) * GeoKeyEntryStruct.size()
            ]
            self.geo_keys.append(GeoKeyEntryStruct.from_buffer(data))

    def record_data_bytes(self):
        b = bytes(self.geo_keys_header)
        b += b"".join(map(bytes, self.geo_keys))
        return b

    def parse_crs(self):
        import pyproj

        # TODO import is done here to avoid cyclic import,
        # this should probably be fixed
        from .geotiff import GeographicTypeGeoKey, ProjectedCSTypeGeoKey

        geographic_cs = None
        projected_cs = None
        for key in self.geo_keys:
            if key.id == ProjectedCSTypeGeoKey.id:
                if 1024 <= key.value_offset <= 32766:
                    # http://docs.opengeospatial.org/is/19-008r4/19-008r4.html#_requirements_class_projectedcrsgeokey
                    # "ProjectedCRSGeoKey values in the range 1024-32766 SHALL be EPSG Projected CRS Codes"
                    projected_cs = pyproj.CRS.from_epsg(key.value_offset)
            elif key.id == GeographicTypeGeoKey.id:
                # http://docs.opengeospatial.org/is/19-008r4/19-008r4.html#_requirements_class_geodeticcrsgeokey
                # GeodeticCRSGeoKey values in the range 1024-32766 SHALL be EPSG geographic 2D or geocentric CRS codes
                if 1024 <= key.value_offset <= 32766:
                    geographic_cs = pyproj.CRS.from_epsg(key.value_offset)

        # Projected Coordinate Systems take precedence since,
        # if they are present, the Geographic CS is probably
        # redundant and the positioning information in the LAS
        # file is projected.
        return projected_cs or geographic_cs

    def __repr__(self):
        return "<{}({} geo_keys)>".format(self.__class__.__name__, len(self.geo_keys))

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (34735,)


class GeoDoubleParamsVlr(BaseKnownVLR):
    """
    Stores all of the `double` valued GeoKeys.
    """

    def __init__(self):
        super().__init__(description="GeoTIFF GeoDoubleParamsTag")
        self.doubles = []

    def parse_record_data(self, record_data):
        sizeof_double = ctypes.sizeof(ctypes.c_double)
        if len(record_data) % sizeof_double != 0:
            raise ValueError(
                "GeoDoubleParams record data length () is not a multiple of sizeof(double) ()".format(
                    len(record_data), sizeof_double
                )
            )
        record_data = bytearray(record_data)
        num_doubles = len(record_data) // sizeof_double
        for i in range(num_doubles):
            b = record_data[i * sizeof_double : (i + 1) * sizeof_double]
            self.doubles.append(ctypes.c_double.from_buffer(b))

    def record_data_bytes(self):
        return b"".join(map(bytes, self.doubles))

    def __repr__(self):
        return "<GeoDoubleParamsVlr({})>".format(self.doubles)

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (34736,)


class GeoAsciiParamsVlr(BaseKnownVLR):
    """
    Stores all of the `ASCII` valued GeoKeys.

    From GeoTIFF's spec:
    To avoid problems with naive tiff dump programs the separator between geokeys is not
    the null-terminator `\0` but `|`.

    """

    def __init__(self):
        super().__init__(description="GeoTIFF GeoAsciiParamsTag")
        self.strings = []

    def parse_record_data(self, record_data):
        self.strings = [s.decode("ascii") for s in record_data.split(NULL_BYTE)]
        self.rd = record_data

    def record_data_bytes(self):
        return NULL_BYTE.join(s.encode("ascii") for s in self.strings)

    def __repr__(self):
        return "<GeoAsciiParamsVlr({})>".format(self.strings)

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (34737,)


class WktMathTransformVlr(BaseKnownVLR):
    """
    From the Spec:
        Note that the math transform WKT record is added for completeness, and a coordinate system WKT
        may or may not require a math transform WKT record

    """

    def __init__(self):
        super().__init__(description="")
        self.string = ""

    def _encode_string(self):
        return encode_to_null_terminated(self.string, codec="utf-8")

    def parse_record_data(self, record_data):
        self.string = record_data.decode("utf-8").rstrip("\0")

    def record_data_bytes(self):
        return self._encode_string()

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (2111,)


class WktCoordinateSystemVlr(BaseKnownVLR):
    """Replaces Coordinates Reference System for new las files (point fmt >= 5)
    "LAS is not using the “ESRI WKT”
    """

    def __init__(self, wkt_string=""):
        super().__init__(description="OGC Transformation Record")
        self.string = wkt_string

    def _encode_string(self):
        return encode_to_null_terminated(self.string, codec="utf-8")

    def parse_record_data(self, record_data):
        self.string = record_data.decode("utf-8").rstrip("\0")

    def record_data_bytes(self):
        return self._encode_string()

    def parse_crs(self):
        import pyproj

        if not self.string:
            return None

        return pyproj.CRS.from_wkt(self.string)

    @staticmethod
    def official_user_id():
        return "LASF_Projection"

    @staticmethod
    def official_record_ids():
        return (2112,)


def vlr_factory(vlr: VLR):
    """Given a vlr tries to find its corresponding KnownVLR class
    that can parse its data.
    If no KnownVLR implementation is found, returns the input vlr unchanged
    """
    user_id = vlr.user_id
    known_vlrs = BaseKnownVLR.__subclasses__()
    for known_vlr in known_vlrs:
        if (
            known_vlr.official_user_id() == user_id
            and vlr.record_id in known_vlr.official_record_ids()
        ):
            try:
                return known_vlr.from_raw(vlr)
            except Exception as err:
                logger.warning(f"Failed to parse {known_vlr}: {err}")
                return vlr

    return vlr



from itertools import takewhile, zip_longest
from typing import Iterable, List, Optional, Type, Union

import numpy as np

from ..errors import LaspyException
from . import dims

Number = Union[np.number, float, int]


class ExtraBytesParams:
    """All parameters needed to create extra bytes"""

    def __init__(
        self,
        name: str,
        type: Union[str, np.dtype, Type[np.uint8]],
        description: str = "",
        offsets: Optional[Iterable[Number]] = None,
        scales: Optional[Iterable[Number]] = None,
        no_data: Optional[Iterable[Number]] = None,
    ) -> None:
        self.name = name
        """ The name of the extra dimension """
        if isinstance(type, str):
            # Work around numpy deprecating support
            # for '1type' strings
            n = "".join(takewhile(lambda c: c.isdigit(), type))
            if n == "1":
                type = type[1:]

        self.type = np.dtype(type)
        """ The type of the extra dimension """
        self.description = description
        """ A description of the extra dimension """
        self.offsets = np.array(offsets) if offsets is not None else offsets
        """ The offsets to use if its a 'scaled dimension', can be none """
        self.scales = np.array(scales) if scales is not None else scales
        """ The scales to use if its a 'scaled dimension', can be none """
        self.no_data = np.array(no_data) if no_data is not None else no_data
        """ The no data values, can be none """


class PointFormat:
    """Class that contains the informations about the dimensions that forms a PointFormat.

    A PointFormat has 'standard' dimensions (dimensions defined in the LAS standard, each
    point format has its set of dimensions), but it can also have extra (non-standard) dimensions
    defined by the user)

    >>> fmt = PointFormat(3)
    >>> all(dim.is_standard for dim in fmt.dimensions)
    True
    >>> dim = fmt.dimension_by_name("classification") # or fmt["classification"]
    >>> dim.max
    31
    >>> dim.min
    0
    >>> dim.num_bits
    5

    """

    def __init__(
        self,
        point_format_id: int,
    ):
        """
        Parameters
        ----------
        point_format_id: int
            point format id
        """
        self.id: int = point_format_id
        self.dimensions: List[dims.DimensionInfo] = []
        composed_dims = dims.COMPOSED_FIELDS[self.id]
        for dim_name in dims.ALL_POINT_FORMATS_DIMENSIONS[self.id]:
            try:
                sub_fields = composed_dims[dim_name]
            except KeyError:
                dimension = dims.DimensionInfo.from_dtype(
                    dim_name, dims.DIMENSIONS_TO_TYPE[dim_name], is_standard=True
                )
                self.dimensions.append(dimension)
            else:
                for sub_field in sub_fields:
                    dimension = dims.DimensionInfo.from_bitmask(
                        sub_field.name, sub_field.mask, is_standard=True
                    )
                    self.dimensions.append(dimension)

    @property
    def standard_dimensions(self) -> Iterable[dims.DimensionInfo]:
        """Returns an iterable of the standard dimensions

        >>> fmt = PointFormat(0)
        >>> standard_dims = list(fmt.standard_dimensions)
        >>> len(standard_dims)
        15
        >>> standard_dims[4].name
        'return_number'


        """
        return (dim for dim in self.dimensions if dim.is_standard)

    @property
    def extra_dimensions(self) -> Iterable[dims.DimensionInfo]:
        return (dim for dim in self.dimensions if dim.is_standard is False)

    @property
    def dimension_names(self) -> Iterable[str]:
        """Returns the names of the dimensions contained in the point format"""
        return (dim.name for dim in self.dimensions)

    @property
    def standard_dimension_names(self) -> Iterable[str]:
        """Returns the names of the extra dimensions in this point format"""
        return (dim.name for dim in self.standard_dimensions)

    @property
    def extra_dimension_names(self) -> Iterable[str]:
        """Returns the names of the extra dimensions in this point format"""
        return (dim.name for dim in self.extra_dimensions)

    @property
    def size(self) -> int:
        """Returns the number of bytes (standard + extra) a point takes

        >>> PointFormat(3).size
        34

        >>> fmt = PointFormat(3)
        >>> fmt.add_extra_dimension(ExtraBytesParams("codification", "uint64"))
        >>> fmt.size
        42
        """
        return int(sum(dim.num_bits for dim in self.dimensions) // 8)

    @property
    def num_standard_bytes(self) -> int:
        """Returns the number of bytes used by standard dims

        >>> fmt = PointFormat(3)
        >>> fmt.add_extra_dimension(ExtraBytesParams("codification", "uint64"))
        >>> fmt.num_standard_bytes
        34
        """
        return int(sum(dim.num_bits for dim in self.standard_dimensions) // 8)

    @property
    def num_extra_bytes(self) -> int:
        """Returns the number of extra bytes

        >>> fmt = PointFormat(3)
        >>> fmt.add_extra_dimension(ExtraBytesParams("codification", "uint64"))
        >>> fmt.num_extra_bytes
        8
        """
        return int(sum(dim.num_bits for dim in self.extra_dimensions) // 8)

    @property
    def has_waveform_packet(self):
        """Returns True if the point format has waveform packet dimensions"""
        dimensions = set(self.dimension_names)
        return all(name in dimensions for name in dims.WAVEFORM_FIELDS_NAMES)

    def dimension_by_name(self, name: str) -> dims.DimensionInfo:
        """Returns the dimension info for the dimension by name

        ValueError is raised if the dimension does not exist un the point format

        >>> info = PointFormat(2).dimension_by_name('number_of_returns')
        >>> info.name == 'number_of_returns'
        True
        >>> info.num_bits == 3
        True


        >>> info = PointFormat(2).dimension_by_name('gps_time')
        Traceback (most recent call last):
        ...
        ValueError: Dimension 'gps_time' does not exist
        """
        for dim in self.dimensions:
            if dim.name == name:
                return dim
        raise ValueError(f"Dimension '{name}' does not exist")

    def add_extra_dimension(self, param: ExtraBytesParams) -> None:
        """Add an extra, user-defined dimension"""
        dim_info = dims.DimensionInfo.from_extra_bytes_param(param)
        # todo: this should be checked in extra bytes param ctor
        if (
            dim_info.num_elements > 3
            and dim_info.kind != dims.DimensionKind.UnsignedInteger
        ):
            raise LaspyException("Extra Dimensions do not support more than 3 elements")
        self.dimensions.append(dim_info)

    def remove_extra_dimension(self, name: str) -> None:
        dimensions = [
            dim for dim in self.dimensions if dim.name == name and not dim.is_standard
        ]

        try:
            dimension = dimensions[0]
        except IndexError:
            if name in self.standard_dimension_names:
                raise LaspyException(
                    f"The dimension named '{name}' is not an extra dimension, "
                    "so it cannot be removed"
                )
            else:
                raise LaspyException(
                    f"'No extra dimension named '{name}' exist"
                ) from None

        self.dimensions = [dim for dim in self.dimensions if dim is not dimension]

    def dtype(self):
        """Returns the numpy.dtype used to store the point records in a numpy array

        .. note::

            The dtype corresponds to the dtype with sub_fields *packed* into their
            composed fields

        """
        dtype = dims.ALL_POINT_FORMATS_DTYPE[self.id]
        descr = dtype.descr
        for extra_dim in self.extra_dimensions:
            descr.append((extra_dim.name, extra_dim.type_str()))
        return np.dtype(descr)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.dimension_by_name(item)
        return self.dimensions[item]

    def __eq__(self, other):
        if self.id != other.id:
            return False

        for my_eb, ot_eb in zip_longest(self.extra_dimensions, other.extra_dimensions):
            if my_eb is None or ot_eb is None:
                return False

            if my_eb != ot_eb:
                return False

        return True

    def __repr__(self):
        return "<PointFormat({}, {} bytes of extra dims)>".format(
            self.id, self.num_extra_bytes
        )


def lost_dimensions(point_fmt_in, point_fmt_out):
    """Returns a list of the names of the dimensions that will be lost
    when converting from point_fmt_in to point_fmt_out
    """

    dimensions_in = set(PointFormat(point_fmt_in).dimension_names)
    dimensions_out = set(PointFormat(point_fmt_out).dimension_names)

    completely_lost = []
    for dim_name in dimensions_in:
        if dim_name not in dimensions_out:
            completely_lost.append(dim_name)
    return completely_lost



from . import errors


class File:
    def __init__(self, *args, **kwargs) -> None:
        raise errors.LaspyException(
            """You are using laspy 2.0, which has several improvements over 1.x
            but with several breaking changes.
            To stay on laspy 1.x: `pip install laspy<2.0.0`
            
            In short:
              - To read a file do: las = laspy.read('somefile.laz')
              - To create a new LAS data do: las = laspy.create(point_format=2, file_version='1.2')
              - To write a file previously read or created: las.write('somepath.las')
            See the documentation for more information about the changes https://laspy.readthedocs.io/en/latest/"""
        )



"""'Entry point' of the library, Contains the various functions meant to be
used directly by a user
"""

import copy
import io
import logging
import os
from pathlib import Path
from typing import Optional, Union

from .compression import DecompressionSelection, LazBackend
from .errors import LaspyException
from .header import LasHeader, Version
from .lasappender import LasAppender
from .lasdata import LasData
from .lasmmap import LasMMAP
from .lasreader import LasReader
from .laswriter import LasWriter
from .point import PointFormat, dims, record
from .vlrs.vlrlist import VLRList

logger = logging.getLogger(__name__)


def open_las(
    source,
    mode="r",
    closefd=True,
    laz_backend=None,
    header=None,
    do_compress=None,
    encoding_errors: str = "strict",
    read_evlrs: bool = True,
    decompression_selection: DecompressionSelection = DecompressionSelection.all(),
) -> Union[LasReader, LasWriter, LasAppender]:
    """The laspy.open opens a LAS/LAZ file in one of the 3 supported
    mode:

     - "r" => Reading => a :class:`laspy.LasReader` will be returned
     - "w" => Writing => a :class:`laspy.LasWriter` will be returned
     - "a" => Appending => a :class:`laspy.LasAppender` will be returned


    When opening a file in 'w' mode, a header (:class:`laspy.LasHeader`)
    is required

        >>> with open_las('tests/data/simple.las') as f:
        ...     print(f.header.point_format.id)
        3


        >>> f = open('tests/data/simple.las', mode='rb')
        >>> with open_las(f,closefd=False) as flas:
        ...     print(flas.header)
        <LasHeader(1.2, <PointFormat(3, 0 bytes of extra dims)>)>
        >>> f.closed
        False
        >>> f.close()
        >>> f.closed
        True


        >>> f = open('tests/data/simple.las', mode='rb')
        >>> with open_las(f) as flas:
        ...    las = flas.read()
        >>> f.closed
        True

    Parameters
    ----------
    source: str or bytes or io.BytesIO
        if source is a str it must be a filename

    mode: Optional, the mode to open the file:
        - "r" for reading (default)
        - "w" for writing
        - "a" for appending

    laz_backend: Optional, laspy.LazBackend, the LAZ backend to use to handle decompression/compression

        By default available backends are detected, see LazBackend to see the
        preference order when multiple backends are available

    header: The header to use when opening in write mode.

    do_compress: optional, bool, only meaningful in writing mode:
        - None (default) guess if compression is needed using the file extension
          or if a laz_backend was explicitely provided
        - True compresses the file
        - False do not compress the file

    closefd: optional, bool, True by default
        Whether the stream/file object shall be closed, this only work
        when using open_las in a with statement. An exception is raised if
        closefd is specified and the source is a filename

    encoding_errors: str, default 'strict'
        Only used in writing and appending mode.
        How encoding errors should be treated.
        Possible values and their explanation can be seen here:
        https://docs.python.org/3/library/codecs.html#error-handlers.

    read_evlrs: bool, default True
            Only applies to 'r' mode.

            If True the evlrs will be read during the __init__ / file opening
            along with the LasHeader.

            It is fine for most of the cases,
            but can be problematic when opening file from a data stream like
            AWS S3 as EVLRs are located at the end of the files, thus
            will require to pull the whole file.

            Does nothing if the input file does not support
            EVLRs

    decompression_selection: DecompressionSelection, default All
        Only applies to 'r' mode and for files which suport selective decompression
        (version >= 1.4 and point format id >= 6), ignored otherwise.

        Allows to select which fields should be decompressed or not, allowing to save time
        by not decompressing unused fields.

        By default all fields are decompressed

    .. versionadded:: 2.4
        The ``read_evlrs`` and ``decompression_selection`` parameters.
    """
    if mode == "r":
        if header is not None:
            raise LaspyException(
                "header argument is not used when opening in read mode, "
                "did you meant to open in write mode ?"
            )
        if do_compress is not None:
            raise LaspyException(
                "do_compress argument is not used when opening in read mode, "
                "did you meant to open in write mode ?"
            )
        if isinstance(source, (str, Path)):
            stream = open(source, mode="rb", closefd=closefd)
        elif isinstance(source, bytes):
            stream = io.BytesIO(source)
        else:
            stream = source
        try:
            return LasReader(
                stream,
                closefd=closefd,
                laz_backend=laz_backend,
                read_evlrs=read_evlrs,
                decompression_selection=decompression_selection,
            )
        except:
            if closefd:
                stream.close()
            raise
    elif mode == "w":
        if header is None:
            raise ValueError("A header is needed when opening a file for writing")

        if isinstance(source, (str, Path)):
            if do_compress is None:
                do_compress = os.path.splitext(source)[1].lower() == ".laz"
            stream = open(source, mode="wb+", closefd=closefd)
        elif isinstance(source, bytes):
            stream = io.BytesIO(source)
        else:
            assert source.seekable()
            stream = source

        try:
            return LasWriter(
                stream,
                header=header,
                do_compress=do_compress,
                laz_backend=laz_backend,
                closefd=closefd,
                encoding_errors=encoding_errors,
            )
        except:
            if closefd:
                stream.close()
            raise
    elif mode == "a":
        if isinstance(source, (str, Path)):
            stream = open(source, mode="rb+", closefd=closefd)
        elif isinstance(source, bytes):
            stream = io.BytesIO(source)
        else:
            stream = source

        try:
            return LasAppender(
                stream,
                closefd=closefd,
                laz_backend=laz_backend,
                encoding_errors=encoding_errors,
            )
        except:
            if closefd:
                stream.close()
            raise
    else:
        raise ValueError(f"Unknown mode '{mode}'")


def read_las(
    source,
    closefd=True,
    laz_backend=LazBackend.detect_available(),
    decompression_selection: DecompressionSelection = DecompressionSelection.all(),
    encoding_errors: str = "strict",
):
    """Entry point for reading las data in laspy

    Reads the whole file into memory.

    >>> las = read_las("tests/data/simple.las")
    >>> las.classification
    <SubFieldView([1 1 1 ... 1 1 1])>

    Parameters
    ----------
    source : str or io.BytesIO
        The source to read data from

    laz_backend: Optional, the backend to use when the file is as LAZ file.
                 By default laspy will find the backend to use by itself.
                 Use if you want a specific backend to be used

    closefd: bool
            if True and the source is a stream, the function will close it
            after it is done reading

    decompression_selection: DecompressionSelection,
        see :func:`laspy.open`

    encoding_errors: str, default 'strict'
        Only used in writing and appending mode.
        How encoding errors should be treated.
        Possible values and their explanation can be seen here:
        https://docs.python.org/3/library/codecs.html#error-handlers.

    Returns
    -------
    laspy.LasData
        The object you can interact with to get access to the LAS points & VLRs


    .. versionadded:: 2.4
        The ``decompression_selection`` parameter.

    .. versionadded:: 2.6
        The ``encoding_errors`` parameter.
    """
    with open_las(
        source,
        closefd=closefd,
        laz_backend=laz_backend,
        decompression_selection=decompression_selection,
        encoding_errors=encoding_errors,
    ) as reader:
        return reader.read()


def mmap_las(filename):
    """MMap a file, much like laspy did"""
    return LasMMAP(filename)


def create_las(
    *,
    point_format: Optional[Union[int, PointFormat]] = None,
    file_version: Optional[Union[str, Version]] = None,
):
    """Function to create a new empty las data object

    .. note::

        If you provide both point_format and file_version
        an exception will be raised if they are not compatible

    >>> las = create_las(point_format=6,file_version="1.2")
    Traceback (most recent call last):
     ...
    laspy.errors.LaspyException: Point format 6 is not compatible with file version 1.2


    If you provide only the point_format the file_version will automatically
    selected for you.

    >>> las = create_las(point_format=0)
    >>> las.header.version == '1.2'
    True

    >>> las = create_las(point_format=PointFormat(6))
    >>> las.header.version == '1.4'
    True


    Parameters
    ----------
    point_format:
        The point format you want the created file to have

    file_version:
        The las version you want the created las to have

    Returns
    -------
    laspy.lasdatas.base.LasBase
       A new las data object

    """
    header = LasHeader(point_format=point_format, version=file_version)
    return LasData(header=header)


def convert(source_las, *, point_format_id=None, file_version=None):
    """Converts a Las from one point format to another
    Automatically upgrades the file version if source file version is not compatible with
    the new point_format_id


    convert to point format 0

    >>> las = read_las('tests/data/simple.las')
    >>> las.header.version
    Version(major=1, minor=2)
    >>> las = convert(las, point_format_id=0)
    >>> las.header.point_format.id
    0
    >>> str(las.header.version)
    '1.2'

    convert to point format 6, which need version >= 1.4
    then convert back to point format 0, version is not downgraded

    >>> las = read_las('tests/data/simple.las')
    >>> str(las.header.version)
    '1.2'
    >>> las = convert(las, point_format_id=6)
    >>> las.header.point_format.id
    6
    >>> str(las.header.version)
    '1.4'
    >>> las = convert(las, point_format_id=0)
    >>> str(las.header.version)
    '1.4'

    an exception is raised if the requested point format is not compatible
    with the file version

    >>> las = read_las('tests/data/simple.las')
    >>> convert(las, point_format_id=6, file_version='1.2')
    Traceback (most recent call last):
     ...
    laspy.errors.LaspyException: Point format 6 is not compatible with file version 1.2

    Parameters
    ----------
    source_las : laspy.lasdatas.base.LasBase
        The source data to be converted

    point_format_id : int, optional
        The new point format id (the default is None, which won't change the source format id)

    file_version : str, optional,
        The new file version. None by default which means that the file_version
        may be upgraded for compatibility with the new point_format. The file version will not
        be downgraded.

    Returns
    -------
        laspy.lasdatas.base.LasBase
    """
    if point_format_id is None:
        point_format_id = source_las.point_format.id

    if file_version is None:
        file_version = max(
            str(source_las.header.version),
            dims.preferred_file_version_for_point_format(point_format_id),
        )
    else:
        file_version = str(file_version)
        dims.raise_if_version_not_compatible_with_fmt(point_format_id, file_version)

    version = Version.from_str(file_version)

    point_format = PointFormat(point_format_id)
    point_format.dimensions.extend(source_las.point_format.extra_dimensions)

    header = copy.deepcopy(source_las.header)
    header.set_version_and_point_format(version, point_format)

    if source_las.evlrs is not None:
        evlrs = VLRList(source_las.evlrs.copy())
    else:
        evlrs = None

    points = record.PackedPointRecord.from_point_record(
        source_las.points, header.point_format
    )
    las = LasData(header=header, points=points)

    if file_version < "1.4" and evlrs is not None and evlrs:
        logger.warning(
            "The source contained {} EVLRs,"
            " they will be lost as version {} doest not support them".format(
                len(evlrs), file_version
            )
        )
    else:
        las.evlrs = evlrs

    return las


def write_then_read_again(
    las, do_compress=False, laz_backend=LazBackend.detect_available()
):
    """writes the given las into memory using BytesIO and
    reads it again, returning the newly read file.

    Mostly used for testing purposes, without having to write to disk
    """
    out = io.BytesIO()
    las.write(out, do_compress=do_compress, laz_backend=laz_backend)
    out.seek(0)
    return read_las(out)



"""The functions related to the LAZ format (compressed LAS)"""

from ._compression.backend import *
from ._compression.format import *

__all__ = [
    "LazBackend",
    "is_point_format_compressed",
    "compressed_id_to_uncompressed",
    "uncompressed_id_to_compressed",
    "DecompressionSelection",
]



import enum


def decompress_and_skip_methods(wrapped_enum):
    """This decorator is to be applied on an enum.IntFlag class

    It will add to this enum some methods:

    - set(self, flag_name) -> Self => to easily set a flag
    - unset(self, flag_name) -> Self => to easily unset a flag
    - is_set(self, flag_name) -> bool => to easily check if a flag is set


    If will also add for each flag member (as long as their name is uppercase)
    the following methods

    - decompress_$flag(self) -> Self
    - skip_$flag(self) -> Self


    # Example

    @load_and_skip_methods
    class Color(IntFlag):
        RED = 1
        GREEN = 2
        BLUE = 4


    will generate (ellipsis to keep it short)

    class Color(IntFlag):
        RED = 1
        GREEN = 2

    def _set(self, mask): ...

    def _unset(self, mask): ...

    def is_set(self, mask): ...


    def decompress_red(self) -> Color: ...
    def decompress_red(self) -> Color : ...
    def is_set_red(self) -> Color : ...
    def load_green(self) -> Color: ...
    def skip_green(self) -> Color: ...
    def is_set_green(self) -> Color : ...
    """

    def set_method(self, mask) -> wrapped_enum:
        return self | mask

    def unset_method(self, mask) -> wrapped_enum:
        return self & (~mask)

    def is_set_method(self, mask) -> bool:
        return (self & mask) != 0

    setattr(wrapped_enum, "_set", set_method)
    setattr(wrapped_enum, "_unset", unset_method)
    setattr(wrapped_enum, "is_set", is_set_method)

    def define_skip_method(mask_member_name: str):
        def skip_method(self):
            return self._unset(getattr(self, mask_member_name))

        return skip_method

    def define_decompress_method(mask_member_name: str):
        def decompress_method(self):
            return self._set(getattr(self, mask_member_name))

        return decompress_method

    def define_is_set_method(mask_member_name: str):
        def is_set_method(self):
            return self.is_set(getattr(self, mask_member_name))

        return is_set_method

    for member in dir(wrapped_enum):
        if not member.isupper():
            continue

        load_name = f"decompress_{member.lower()}"
        skip_name = f"skip_{member.lower()}"
        is_set_name = f"is_set_{member.lower()}"

        setattr(wrapped_enum, load_name, define_decompress_method(member))
        setattr(wrapped_enum, skip_name, define_skip_method(member))
        setattr(wrapped_enum, is_set_name, define_is_set_method(member))

    return wrapped_enum


@decompress_and_skip_methods
class DecompressionSelection(enum.IntFlag):
    """
    Holds which fields to decompress  or not.

    Only used for files with version >= 1.4 && point format id >= 6.

    Ignored on other cases.


    Each flag in the enum has a corresponding ``decompress_$name`` and
    ``skip_$name`` methods to easily create a selection.

    >>> import laspy
    >>> # Creating a selection that decompresses the base + z field
    >>> selection = laspy.DecompressionSelection.base().decompress_z()
    >>> selection.is_set(laspy.DecompressionSelection.Z)
    True
    >>> selection.is_set(laspy.DecompressionSelection.INTENSITY)
    False

    >>> # Creating a selection that decompresses all fields but the intensity
    >>> selection = laspy.DecompressionSelection.all().skip_intensity()
    >>> selection.is_set(laspy.DecompressionSelection.INTENSITY)
    False
    >>> selection.is_set(laspy.DecompressionSelection.Z)
    True


    .. versionadded:: 2.4

    """

    #: Flag to decompress x, y, return number, number of returns and scanner channel
    XY_RETURNS_CHANNEL = enum.auto()
    #: Flag to decompress z
    Z = enum.auto()
    #: Flag to decompress the classification
    CLASSIFICATION = enum.auto()
    #: Flag to decompress the classification flags (withheld, key point, overlap, etc)
    FLAGS = enum.auto()
    #: Flag to decompress the intensity
    INTENSITY = enum.auto()
    #: Flag to decompress the scan angle
    SCAN_ANGLE = enum.auto()
    #: Flag to decompress the user data
    USER_DATA = enum.auto()
    #: Flag to decompress the point source id
    POINT_SOURCE_ID = enum.auto()
    #: Flag to decompress the gps time
    GPS_TIME = enum.auto()
    #: Flag to decompress the red, green, blue
    RGB = enum.auto()
    #: Flag to decompress the nir
    NIR = enum.auto()
    #: Flag to decompress the wavepacket
    WAVEPACKET = enum.auto()
    #: Flag to decompress all the extra bytes
    ALL_EXTRA_BYTES = enum.auto()

    @classmethod
    def all(cls) -> "DecompressionSelection":
        """Returns a selection where all fields will be decompressed"""

        selection = cls.base()
        for flag in cls:
            selection = selection._set(flag)

        return selection

    @classmethod
    def base(cls) -> "DecompressionSelection":
        """
        Returns a decompression selection where only the base
        x, y, return number, number of returns and scanner channel will be decompressed
        """
        return cls.xy_returns_channel()

    @classmethod
    def xy_returns_channel(cls) -> "DecompressionSelection":
        """
        Returns a decompression selection where only the base
        x, y, return number, number of returns and scanner channel will be decompressed
        """
        return cls.XY_RETURNS_CHANNEL

    def to_lazrs(self) -> "lazrs.DecompressionSelection":
        import lazrs

        variant_mapping = {
            DecompressionSelection.XY_RETURNS_CHANNEL: lazrs.SELECTIVE_DECOMPRESS_XY_RETURNS_CHANNEL,
            DecompressionSelection.Z: lazrs.SELECTIVE_DECOMPRESS_Z,
            DecompressionSelection.CLASSIFICATION: lazrs.SELECTIVE_DECOMPRESS_CLASSIFICATION,
            DecompressionSelection.FLAGS: lazrs.SELECTIVE_DECOMPRESS_FLAGS,
            DecompressionSelection.INTENSITY: lazrs.SELECTIVE_DECOMPRESS_INTENSITY,
            DecompressionSelection.SCAN_ANGLE: lazrs.SELECTIVE_DECOMPRESS_SCAN_ANGLE,
            DecompressionSelection.USER_DATA: lazrs.SELECTIVE_DECOMPRESS_USER_DATA,
            DecompressionSelection.POINT_SOURCE_ID: lazrs.SELECTIVE_DECOMPRESS_POINT_SOURCE_ID,
            DecompressionSelection.GPS_TIME: lazrs.SELECTIVE_DECOMPRESS_GPS_TIME,
            DecompressionSelection.RGB: lazrs.SELECTIVE_DECOMPRESS_RGB,
            DecompressionSelection.NIR: lazrs.SELECTIVE_DECOMPRESS_NIR,
            DecompressionSelection.WAVEPACKET: lazrs.SELECTIVE_DECOMPRESS_WAVEPACKET,
            DecompressionSelection.ALL_EXTRA_BYTES: lazrs.SELECTIVE_DECOMPRESS_ALL_EXTRA_BYTES,
        }
        lazrs_selection = lazrs.SELECTIVE_DECOMPRESS_XY_RETURNS_CHANNEL
        for variant in DecompressionSelection:
            lazrs_selection |= variant_mapping[variant] if self.is_set(variant) else 0

        return lazrs.DecompressionSelection(lazrs_selection)

    def to_laszip(self) -> int:
        import laszip

        variant_mapping = {
            DecompressionSelection.XY_RETURNS_CHANNEL: laszip.DECOMPRESS_SELECTIVE_CHANNEL_RETURNS_XY,
            DecompressionSelection.Z: laszip.DECOMPRESS_SELECTIVE_Z,
            DecompressionSelection.CLASSIFICATION: laszip.DECOMPRESS_SELECTIVE_CLASSIFICATION,
            DecompressionSelection.FLAGS: laszip.DECOMPRESS_SELECTIVE_FLAGS,
            DecompressionSelection.INTENSITY: laszip.DECOMPRESS_SELECTIVE_INTENSITY,
            DecompressionSelection.SCAN_ANGLE: laszip.DECOMPRESS_SELECTIVE_SCAN_ANGLE,
            DecompressionSelection.USER_DATA: laszip.DECOMPRESS_SELECTIVE_USER_DATA,
            DecompressionSelection.POINT_SOURCE_ID: laszip.DECOMPRESS_SELECTIVE_POINT_SOURCE,
            DecompressionSelection.GPS_TIME: laszip.DECOMPRESS_SELECTIVE_GPS_TIME,
            DecompressionSelection.RGB: laszip.DECOMPRESS_SELECTIVE_RGB,
            DecompressionSelection.NIR: laszip.DECOMPRESS_SELECTIVE_NIR,
            DecompressionSelection.WAVEPACKET: laszip.DECOMPRESS_SELECTIVE_WAVEPACKET,
            DecompressionSelection.ALL_EXTRA_BYTES: laszip.DECOMPRESS_SELECTIVE_EXTRA_BYTES,
        }
        laszip_selection = laszip.DECOMPRESS_SELECTIVE_CHANNEL_RETURNS_XY
        for variant in DecompressionSelection:
            laszip_selection |= variant_mapping[variant] if self.is_set(variant) else 0

        return laszip_selection



"""This module contains things like the definitions of the point formats dimensions,
the mapping between dimension names and their type, mapping between point format and
compatible file version
"""

import abc
import collections
import operator
from collections import UserDict
from enum import Enum
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np

from .. import errors
from . import packing

ValueType = TypeVar("ValueType")


class PointFormatDict(UserDict, Generic[ValueType]):
    """Simple wrapper around a dict that changes
    the exception raised when accessing a key that is not-present

    """

    def __init__(self, wrapped_dict: Dict[int, ValueType]):
        super().__init__(wrapped_dict)

    def __getitem__(self, key: int) -> ValueType:
        try:
            return self.data[key]
        except KeyError:
            raise errors.PointFormatNotSupported(key) from None


class SubField(NamedTuple):
    name: str
    mask: int


def _point_format_to_dtype(
    point_format: Iterable[str], dimensions_to_type: Mapping[str, np.dtype]
) -> np.dtype:
    """build the numpy.dtype for a point format

    Parameters:
    ----------
    point_format : iterable of str
        The dimensions names of the point format
    dimensions : dict
        The dictionary of dimensions
    Returns
    -------
    numpy.dtype
        The dtype for the input point format
    """
    return np.dtype(
        [(dim_name, dimensions_to_type[dim_name]) for dim_name in point_format]
    )


def _build_point_formats_dtypes(
    point_format_dimensions: Mapping[int, Tuple[str]],
    dimensions_dict: Mapping[str, np.dtype],
) -> Dict[int, np.dtype]:
    """Builds the dict mapping point format id to numpy.dtype
    In the dtypes, bit fields are still packed, and need to be unpacked each time
    you want to access them
    """
    return {
        fmt_id: _point_format_to_dtype(point_fmt, dimensions_dict)
        for fmt_id, point_fmt in point_format_dimensions.items()
    }


OLD_LASPY_NAMES = {
    "flag_byte": "bit_fields",
    "return_num": "return_number",
    "num_returns": "number_of_returns",
    "scan_dir_flag": "scan_direction_flag",
    "edge_flight_line": "edge_of_flight_line",
    "pt_src_id": "point_source_id",
    "wave_packet_desc_index": "wavepacket_index",
    "byte_offset_to_waveform_data": "wavepacket_offset",
    "waveform_packet_size": "wavepacket_size",
    "return_point_waveform_loc": "return_point_wave_location",
}

# Definition of the points dimensions and formats
# LAS version [1.0, 1.1, 1.2, 1.3, 1.4]
DIMENSIONS_TO_TYPE: Dict[str, np.dtype] = {
    "X": np.dtype("i4"),
    "Y": np.dtype("i4"),
    "Z": np.dtype("i4"),
    "intensity": np.dtype("u2"),
    "bit_fields": np.dtype("u1"),
    "raw_classification": np.dtype("u1"),
    "scan_angle_rank": np.dtype("i1"),
    "user_data": np.dtype("u1"),
    "point_source_id": np.dtype("u2"),
    "gps_time": np.dtype("f8"),
    "red": np.dtype("u2"),
    "green": np.dtype("u2"),
    "blue": np.dtype("u2"),
    # Waveform related dimensions
    "wavepacket_index": np.dtype("u1"),
    "wavepacket_offset": np.dtype("u8"),
    "wavepacket_size": np.dtype("u4"),
    "return_point_wave_location": np.dtype("f4"),
    "x_t": np.dtype("f4"),
    "y_t": np.dtype("f4"),
    "z_t": np.dtype("f4"),
    # Las 1.4
    "classification_flags": np.dtype("u1"),
    "scan_angle": np.dtype("i2"),
    "classification": np.dtype("u1"),
    "nir": np.dtype("u2"),
}

POINT_FORMAT_0: Tuple[str, ...] = (
    "X",
    "Y",
    "Z",
    "intensity",
    "bit_fields",
    "raw_classification",
    "scan_angle_rank",
    "user_data",
    "point_source_id",
)

POINT_FORMAT_6: Tuple[str, ...] = (
    "X",
    "Y",
    "Z",
    "intensity",
    "bit_fields",
    "classification_flags",
    "classification",
    "user_data",
    "scan_angle",
    "point_source_id",
    "gps_time",
)

WAVEFORM_FIELDS_NAMES: Tuple[str, ...] = (
    "wavepacket_index",
    "wavepacket_offset",
    "wavepacket_size",
    "return_point_wave_location",
    "x_t",
    "y_t",
    "z_t",
)

COLOR_FIELDS_NAMES: Tuple[str, ...] = ("red", "green", "blue")

POINT_FORMAT_DIMENSIONS = PointFormatDict(
    {
        0: POINT_FORMAT_0,
        1: POINT_FORMAT_0 + ("gps_time",),
        2: POINT_FORMAT_0 + COLOR_FIELDS_NAMES,
        3: POINT_FORMAT_0 + ("gps_time",) + COLOR_FIELDS_NAMES,
        4: POINT_FORMAT_0 + ("gps_time",) + WAVEFORM_FIELDS_NAMES,
        5: POINT_FORMAT_0 + ("gps_time",) + COLOR_FIELDS_NAMES + WAVEFORM_FIELDS_NAMES,
        6: POINT_FORMAT_6,
        7: POINT_FORMAT_6 + COLOR_FIELDS_NAMES,
        8: POINT_FORMAT_6 + COLOR_FIELDS_NAMES + ("nir",),
        9: POINT_FORMAT_6 + WAVEFORM_FIELDS_NAMES,
        10: POINT_FORMAT_6 + COLOR_FIELDS_NAMES + ("nir",) + WAVEFORM_FIELDS_NAMES,
    }
)

# sub fields of the 'bit_fields' dimension
RETURN_NUMBER_MASK_0 = 0b00000111
NUMBER_OF_RETURNS_MASK_0 = 0b00111000
SCAN_DIRECTION_FLAG_MASK_0 = 0b01000000
EDGE_OF_FLIGHT_LINE_MASK_0 = 0b10000000

# sub fields of the 'raw_classification' dimension
CLASSIFICATION_MASK_0 = 0b00011111
SYNTHETIC_MASK_0 = 0b00100000
KEY_POINT_MASK_0 = 0b01000000
WITHHELD_MASK_0 = 0b10000000

# sub fields of the bit_fields
RETURN_NUMBER_MASK_6 = 0b00001111
NUMBER_OF_RETURNS_MASK_6 = 0b11110000

# sub fields of classification flags
CLASSIFICATION_FLAGS_MASK_6 = 0b00001111

SYNTHETIC_MASK_6 = 0b00000001
KEY_POINT_MASK_6 = 0b00000010
WITHHELD_MASK_6 = 0b00000100
OVERLAP_MASK_6 = 0b00001000
SCANNER_CHANNEL_MASK_6 = 0b00110000
SCAN_DIRECTION_FLAG_MASK_6 = 0b01000000
EDGE_OF_FLIGHT_LINE_MASK_6 = 0b10000000

COMPOSED_FIELDS_0: Dict[str, List[SubField]] = {
    "bit_fields": [
        SubField("return_number", RETURN_NUMBER_MASK_0),
        SubField("number_of_returns", NUMBER_OF_RETURNS_MASK_0),
        SubField("scan_direction_flag", SCAN_DIRECTION_FLAG_MASK_0),
        SubField("edge_of_flight_line", EDGE_OF_FLIGHT_LINE_MASK_0),
    ],
    "raw_classification": [
        SubField("classification", CLASSIFICATION_MASK_0),
        SubField("synthetic", SYNTHETIC_MASK_0),
        SubField("key_point", KEY_POINT_MASK_0),
        SubField("withheld", WITHHELD_MASK_0),
    ],
}

COMPOSED_FIELDS_6: Dict[str, List[SubField]] = {
    "bit_fields": [
        SubField("return_number", RETURN_NUMBER_MASK_6),
        SubField("number_of_returns", NUMBER_OF_RETURNS_MASK_6),
    ],
    "classification_flags": [
        SubField("synthetic", SYNTHETIC_MASK_6),
        SubField("key_point", KEY_POINT_MASK_6),
        SubField("withheld", WITHHELD_MASK_6),
        SubField("overlap", OVERLAP_MASK_6),
        SubField("scanner_channel", SCANNER_CHANNEL_MASK_6),
        SubField("scan_direction_flag", SCAN_DIRECTION_FLAG_MASK_6),
        SubField("edge_of_flight_line", EDGE_OF_FLIGHT_LINE_MASK_6),
    ],
}

# Dict giving the composed fields for each point_format_id
COMPOSED_FIELDS = PointFormatDict(
    {
        0: COMPOSED_FIELDS_0,
        1: COMPOSED_FIELDS_0,
        2: COMPOSED_FIELDS_0,
        3: COMPOSED_FIELDS_0,
        4: COMPOSED_FIELDS_0,
        5: COMPOSED_FIELDS_0,
        6: COMPOSED_FIELDS_6,
        7: COMPOSED_FIELDS_6,
        8: COMPOSED_FIELDS_6,
        9: COMPOSED_FIELDS_6,
        10: COMPOSED_FIELDS_6,
    }
)

VERSION_TO_POINT_FMT: Dict[str, Tuple[int, ...]] = {
    "1.1": (0, 1),
    "1.2": (0, 1, 2, 3),
    "1.3": (0, 1, 2, 3, 4, 5),
    "1.4": (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
    "1.5": (6, 7, 8, 9, 10),
}

POINT_FORMATS_DTYPE = PointFormatDict(
    _build_point_formats_dtypes(POINT_FORMAT_DIMENSIONS, DIMENSIONS_TO_TYPE)
)
# This Dict maps point_format_ids to their dimensions names
ALL_POINT_FORMATS_DIMENSIONS = PointFormatDict({**POINT_FORMAT_DIMENSIONS})
# This Dict maps point_format_ids to their numpy.dtype
# the dtype corresponds to the de packed data
ALL_POINT_FORMATS_DTYPE = PointFormatDict({**POINT_FORMATS_DTYPE})


def get_sub_fields_dict(point_format_id: int) -> Dict[str, Tuple[str, SubField]]:
    sub_fields_dict = {}
    for composed_dim_name, sub_fields in COMPOSED_FIELDS[point_format_id].items():
        for sub_field in sub_fields:
            sub_fields_dict[sub_field.name] = (composed_dim_name, sub_field)
    return sub_fields_dict


class DimensionKind(Enum):
    SignedInteger = 0
    UnsignedInteger = 1
    FloatingPoint = 2
    BitField = 3

    @classmethod
    def from_letter(cls, letter: str) -> "DimensionKind":
        if letter == "u":
            return cls.UnsignedInteger
        elif letter == "i":
            return cls.SignedInteger
        elif letter == "f":
            return cls.FloatingPoint
        else:
            raise ValueError(f"Unknown type letter '{letter}'")

    def letter(self) -> Optional[str]:
        if self == DimensionKind.UnsignedInteger:
            return "u"
        elif self == DimensionKind.SignedInteger:
            return "i"
        elif self == DimensionKind.FloatingPoint:
            return "f"
        else:
            return None


def num_bit_set(n: int) -> int:
    """Count the number of bits that are set (1) in the number n

    Brian Kernighan's algorithm
    """
    count = 0
    while n != 0:
        count += 1
        n = n & (n - 1)
    return count


class DimensionInfo(NamedTuple):
    """Tuple that contains information of a dimension"""

    name: str
    kind: DimensionKind
    num_bits: int
    num_elements: int = 1
    is_standard: bool = True
    description: str = ""
    offsets: Optional[np.ndarray] = None
    scales: Optional[np.ndarray] = None
    no_data: Optional[np.ndarray] = None

    @classmethod
    def from_extra_bytes_param(cls, params):
        me = cls(
            params.name,
            DimensionKind.from_letter(params.type.base.kind),
            params.type.itemsize * 8,
            params.type.shape[0] if params.type.ndim == 1 else 1,
            False,
            params.description,
            params.offsets,
            params.scales,
            params.no_data,
        )
        me._validate()
        return me

    @classmethod
    def from_dtype(
        cls,
        name: str,
        dtype: np.dtype,
        is_standard: bool = True,
        description: str = "",
        offsets: Optional[np.ndarray] = None,
        scales: Optional[np.ndarray] = None,
    ) -> "DimensionInfo":
        if dtype.ndim != 0:
            num_elements = dtype.shape[0]
        else:
            num_elements = 1

        kind = DimensionKind.from_letter(dtype.base.kind)
        num_bits = dtype.itemsize * 8

        self = cls(
            name,
            kind,
            num_bits,
            num_elements,
            is_standard,
            description=description,
            offsets=offsets,
            scales=scales,
        )
        self._validate()
        return self

    @classmethod
    def from_bitmask(
        cls, name: str, bit_mask: int, is_standard: bool = False
    ) -> "DimensionInfo":
        kind = DimensionKind.BitField
        bit_size = num_bit_set(bit_mask)
        return cls(name, kind, bit_size, is_standard=is_standard)

    @property
    def num_bytes(self) -> int:
        return int(self.num_bits // 8)

    @property
    def num_bytes_singular_element(self) -> int:
        return int(self.num_bits // (8 * self.num_elements))

    @property
    def is_scaled(self) -> bool:
        return self.scales is not None or self.offsets is not None

    @property
    def max(self):
        if self.kind == DimensionKind.BitField:
            return (2**self.num_bits) - 1
        elif self.kind == DimensionKind.FloatingPoint:
            return np.finfo(self.type_str()).max
        else:
            return np.iinfo(self.type_str()).max

    @property
    def min(self):
        if (
            self.kind == DimensionKind.BitField
            or self.kind == DimensionKind.UnsignedInteger
        ):
            return 0
        elif self.kind == DimensionKind.FloatingPoint:
            return np.finfo(self.type_str()).min
        else:
            return np.iinfo(self.type_str()).min

    def type_str(self) -> Optional[str]:
        if self.kind == DimensionKind.BitField:
            return None

        if self.num_elements == 1:
            return f"{self.kind.letter()}{self.num_bytes_singular_element}"
        return (
            f"{self.num_elements}{self.kind.letter()}{self.num_bytes_singular_element}"
        )

    @property
    def dtype(self) -> Optional[np.dtype]:
        type_str = self.type_str()
        if type_str is not None:
            return np.dtype(type_str)
        return None

    def __eq__(self, other: "DimensionInfo") -> bool:
        # Named Tuple implements that for us, but
        # when scales and offset are not None (thus are array)
        # The default '==' won't work
        # (ValueError, value of an array with more than one element is ambiguous)
        return (
            self.name == other.name
            and self.kind == other.kind
            and self.num_bits == other.num_bits
            and self.is_standard == other.is_standard
            and self.description == other.description
            and np.all(self.offsets == other.offsets)
            and np.all(self.scales == other.scales)
        )

    def __ne__(self, other: "DimensionInfo") -> bool:
        return not self == other

    def _validate(self):
        if (self.offsets is not None and self.scales is None) or (
            self.offsets is None and self.scales is not None
        ):
            raise ValueError("Cannot provide scales without offsets and vice-versa")

        if self.offsets is not None and len(self.offsets) != self.num_elements:
            raise ValueError(
                f"len(offsets) ({len(self.offsets)}) is not the same as the number of elements ({self.num_elements})"
            )

        if self.scales is not None and len(self.scales) != self.num_elements:
            raise ValueError(
                f"len(scales) ({len(self.scales)}) is not the same as the number of elements ({self.num_elements})"
            )

        if self.no_data is not None and len(self.no_data) != self.num_elements:
            raise ValueError(
                f"len(no_data) ({len(self.no_data)}) is not the same as the number of elements ({self.num_elements})"
            )


def size_of_point_format_id(point_format_id: int) -> int:
    return ALL_POINT_FORMATS_DTYPE[point_format_id].itemsize


def preferred_file_version_for_point_format(point_format_id: int) -> str:
    def inclusive_range(start: int, stop: int):
        return range(start, stop + 1)

    if point_format_id in inclusive_range(0, 3):
        return "1.2"
    elif point_format_id in inclusive_range(4, 5):
        return "1.3"
    elif point_format_id in inclusive_range(6, 10):
        return "1.4"
    else:
        raise errors.PointFormatNotSupported(point_format_id)


def min_point_format_for_version(version: str) -> int:
    return VERSION_TO_POINT_FMT[version][0]


def supported_versions() -> Set[str]:
    """Returns the set of supported file versions"""
    return set(VERSION_TO_POINT_FMT.keys())


def supported_point_formats() -> Set[int]:
    """Returns a set of all the point formats supported in laspy"""
    return set(POINT_FORMAT_DIMENSIONS.keys())


def is_point_fmt_compatible_with_version(
    point_format_id: int, file_version: str
) -> bool:
    """Returns true if the file version support the point_format_id"""
    try:
        return point_format_id in VERSION_TO_POINT_FMT[str(file_version)]
    except KeyError:
        raise errors.FileVersionNotSupported(file_version)


def raise_if_version_not_compatible_with_fmt(point_format_id: int, file_version: str):
    if not is_point_fmt_compatible_with_version(point_format_id, file_version):
        raise errors.LaspyException(
            "Point format {} is not compatible with file version {}".format(
                point_format_id, file_version
            )
        )


def _convert_array_views_to_array(
    view_class: Type, some_args: Union[List[Any], Tuple[Any, ...]]
) -> List[Any]:
    converted_args = []
    for arg in some_args:
        if isinstance(arg, (list, tuple)):
            converted_args.append(_convert_array_views_to_array(view_class, arg))
        elif isinstance(arg, view_class):
            converted_args.append(np.array(arg))
        else:
            converted_args.append(arg)

    return converted_args


class ArrayView(abc.ABC):
    def __init__(self, array) -> None:
        self.array = array

    @abc.abstractmethod
    def __array__(self, *args, **kwargs) -> np.ndarray: ...

    @abc.abstractmethod
    def __getitem__(self, item): ...

    @abc.abstractmethod
    def __setitem__(self, key, value): ...

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        inpts = _convert_array_views_to_array(self.__class__, inputs)
        return getattr(ufunc, method)(*inpts, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        argslist = _convert_array_views_to_array(self.__class__, args)
        return func(*argslist, **kwargs)

    def copy(self) -> np.ndarray:
        return np.array(self)

    @property
    def dtype(self):
        return self.array.dtype

    @property
    def shape(self):
        return self.array.shape

    @property
    def ndim(self):
        return self.array.ndim

    def max(self, *args, **kwargs):
        return np.array(self).max(*args, **kwargs)

    def min(self, *args, **kwargs):
        return np.array(self).min(*args, **kwargs)

    def __len__(self):
        return len(self.array)

    def __lt__(self, other):
        return np.array(self) < other

    def __le__(self, other):
        return np.array(self) <= other

    def __ge__(self, other):
        return np.array(self) >= other

    def __gt__(self, other):
        return np.array(self) > other

    def __eq__(self, other):
        return np.array(self) == other

    def __ne__(self, other):
        return np.array(self) != other

    def __add__(self, other):
        return np.array(self) + other

    def __sub__(self, other):
        return np.array(self) - other

    def __mul__(self, other):
        return np.array(self) * other

    def __truediv__(self, other):
        return np.array(self) / other

    def __floordiv__(self, other):
        return np.array(self) // other

    def __repr__(self):
        return f"<{self.__class__.__name__}({np.array(self)})>"


class SubFieldView(ArrayView):
    """Offers a view onto a LAS field that is a bit field.

    This class allows to read and modify, the array that stores the
    bit field directly.
    """

    def __init__(self, array: np.ndarray, bit_mask):
        super().__init__(array)
        self.bit_mask = self.array.dtype.type(bit_mask)
        self.lsb = packing.least_significant_bit_set(bit_mask)
        self.max_value_allowed = int(self.bit_mask >> self.lsb)

    def masked_array(self):
        return (self.array & self.bit_mask) >> self.lsb

    def _do_comparison(self, value, comp):
        if isinstance(value, (int, type(self.array.dtype))):
            if value > self.max_value_allowed:
                return np.zeros_like(self.array, bool)
        return comp(self.array & self.bit_mask, value << self.lsb)

    def __array__(self, *args, **kwargs):
        ret = self.masked_array()
        if not isinstance(ret, np.ndarray):
            ret = np.array(ret)
        return ret

    def __lt__(self, other):
        return self._do_comparison(other, operator.lt)

    def __le__(self, other):
        return self._do_comparison(other, operator.le)

    def __ge__(self, other):
        return self._do_comparison(other, operator.ge)

    def __gt__(self, other):
        return self._do_comparison(other, operator.gt)

    def __setitem__(self, key, value):
        # bail out on empty sequences
        if isinstance(value, collections.abc.Sized) and len(value) == 0:
            return

        if np.max(value) > self.max_value_allowed:
            raise OverflowError(
                f"value {np.max(value)} is greater than allowed (max: {self.max_value_allowed})"
            )
        value = np.asarray(value)
        self.array[key] &= ~self.bit_mask

        # This is not allowed without a casting="unsafe" argument
        # in Numpy 2.0
        # self.array[key] |= shifted
        shifted = value << self.lsb
        self.array[key] = np.bitwise_or(self.array[key], shifted, casting="unsafe")

    def __getitem__(self, item):
        sliced = SubFieldView(self.array[item], int(self.bit_mask))
        if isinstance(item, int):
            return sliced.masked_array()
        return sliced


class ScaledArrayView(ArrayView):
    def __init__(
        self,
        array: np.ndarray,
        scale: Union[float, np.ndarray],
        offset: Union[float, np.ndarray],
    ) -> None:
        super().__init__(array)
        self.scale = scale
        self.offset = offset

    def scaled_array(self):
        return self._apply_scale(self.array)

    def __array__(self, *args, **kwargs) -> np.ndarray:
        return self.scaled_array()

    def _apply_scale(self, value):
        return (value * self.scale) + self.offset

    def _remove_scale(self, value, sub=None):
        if sub is None:
            return np.round((value - self.offset) / self.scale)
        else:
            return np.round((value - self.offset[sub]) / self.scale[sub])

    def max(self, *args, **kwargs):
        return self._apply_scale(self.array.max(*args, **kwargs))

    def min(self, *args, **kwargs):
        return self._apply_scale(self.array.min(*args, **kwargs))

    @property
    def dtype(self):
        return np.dtype(np.float64)

    def _do_comparison(self, other, op):
        # The implementation of comparison by the base class
        # does a conversion to np.array of self, which is not free
        # we try to avoid that here
        if isinstance(other, (int, float, np.number)):
            other = self._remove_scale(other)
            return getattr(self.array, op)(other)
        return getattr(super(), op)(other)

    def __ge__(self, other):
        return self._do_comparison(other, "__ge__")

    def __gt__(self, other):
        return self._do_comparison(other, "__gt__")

    def __le__(self, other):
        return self._do_comparison(other, "__le__")

    def __lt__(self, other):
        return self._do_comparison(other, "__lt__")

    def __eq__(self, other):
        return self._do_comparison(other, "__eq__")

    def __ne__(self, other):
        return self._do_comparison(other, "__ne__")

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._apply_scale(self.array[item])
        elif isinstance(item, slice):
            return self.__class__(self.array[item], self.scale, self.offset)
        else:
            sliced_array = self.array[item]
            if len(item) == 2:
                if item[1] is Ellipsis:
                    # item is (index, ...), it queries for all the dimensions
                    # of a point or set of point, so we don't slice the scales/offsets
                    return self.__class__(sliced_array, self.scale, self.offset)
                else:
                    # item is something like (something, index)
                    # it queries for one dimension or set of dimension
                    # for all the points, so we need to slice the scales/offsets
                    return self.__class__(
                        sliced_array, self.scale[item[1]], self.offset[item[1]]
                    )
            return self.__class__(sliced_array, self.scale, self.offset)

    def __setitem__(self, key, value):
        # bail out on empty sequences
        if isinstance(value, collections.abc.Sized) and len(value) == 0:
            return

        try:
            info = np.iinfo(self.array.dtype)
        except ValueError:
            info = np.finfo(self.array.dtype)

        new_max = np.max(value)
        new_min = np.min(value)

        max_allowed = self._apply_scale(info.max)
        min_allowed = self._apply_scale(info.min)

        if np.any(new_max > max_allowed) or np.any(new_min < min_allowed):
            raise OverflowError(
                "Values given do not fit after applying offset and scale"
            )
        if isinstance(value, ScaledArrayView):
            value = np.array(value)

        try:
            key_len = len(key)
        except TypeError:
            key_len = 1

        if key_len == 2:
            self.array[key] = self._remove_scale(value, sub=key[1])
        else:
            self.array[key] = self._remove_scale(value)

    def __repr__(self):
        return f"<ScaledArrayView({self.scaled_array()})>"



import io
import mmap

from . import lasdata
from .header import LasHeader
from .point import record
from .typehints import PathLike

WHOLE_FILE = 0


class LasMMAP(lasdata.LasData):
    """Memory map a LAS file.
    It works like a regular LasData however the data is not actually read in memory,

    Access to dimensions are made directly from the file itself, changes made to the points
    are directly reflected in the mmap file.

    Vlrs cannot be modified.

    This can be useful if you want to be able to process a big LAS file

    .. note::
        A LAZ (compressed LAS) cannot be mmapped
    """

    def __init__(self, filename: PathLike) -> None:
        fileref = open(filename, mode="r+b")

        m = mmap.mmap(fileref.fileno(), length=WHOLE_FILE, access=mmap.ACCESS_WRITE)
        header = LasHeader.read_from(m)
        if header.are_points_compressed:
            raise ValueError("Cannot mmap a compressed LAZ file")

        points_data = record.PackedPointRecord.from_buffer(
            m,
            header.point_format,
            count=header.point_count,
            offset=header.offset_to_point_data,
        )
        super().__init__(header=header, points=points_data)

        self.fileref, self.mmap = fileref, m
        self.mmap.seek(0, io.SEEK_SET)

    def close(self) -> None:
        # These need to be set to None, so that
        # mmap.close() does not give an error because
        # there are still exported pointers
        self._points = None
        self.mmap.close()
        self.fileref.close()

    def __enter__(self) -> "LasMMAP":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()
